"""Various analysis functions."""
import os
import numpy as np
import pandas as pd
import xarray as xr
from scipy.stats import t
from uncertainties.unumpy import uarray


def get_band_spectrum(med, region=None, component_mng_names=None,
                      comp_time_slices={}, **kwargs):
    """Get band spectrum, i.e. the integration of power spectrum over
    frequency bands.

    :param med: Mediator.
    :param region: Region to select, `global` or `None`.
      Default is `None`, in which case all regions are kept.
      If `'global'`, data is aggregated over regions.
    :param component_mng_names: Names of components to select.
      Default is `None`, in which case all components in mediator are selected.
    :param comp_time_slices: Time slice per component.
    :type med: :py:class:`.mediator.Mediator`
    :type region: str
    :type component_mng_names: collection
    :type comp_time_slices: :py:class:`slice`-valued mutable

    :returns: Original and filtered datasets per component and result.
    :rtype: :py:class:`tuple` of mappings
    """
    freqs = ['Y', 'D', 'H']
    ds = {}
    var_tot = {}
    filt = {}
    var_filt = {}

    # Read hourly capacity factors from ENTSO-E and GSE
    component_mng_names = component_mng_names or med.component_managers.keys()
    for component_mng_name in component_mng_names:
        # Get result
        component_mng = med.component_managers[component_mng_name]
        file_dir = med.cfg.get_project_data_directory(component_mng, **kwargs)
        time_slice = (None if comp_time_slices is None else
                      comp_time_slices.get(component_mng_name))

        ds[component_mng_name] = {}
        var_tot[component_mng_name] = {}
        filt[component_mng_name] = {}
        var_filt[component_mng_name] = {}
        for result_mng_name, result_mng in (
                component_mng.result_managers.items()):
            # Get array for result and component
            da = result_mng.get_data()[result_mng_name]
            try:
                da = da.sel(component=result_mng.result_name)
            except ValueError:
                pass

            # Select time slice if needed
            if time_slice is not None:
                da = da.sel(time=time_slice)

            coords = [('frequency', freqs)]
            coords_shape = (len(freqs),)
            if region is not None:
                # Select region
                da = (da.sum('region', keep_attrs=True) if region == 'global'
                      else da.sel(region=region))
            else:
                coords.append(da.coords['region'])
                coords_shape += (len(da.coords['region']),)
            ds[component_mng_name][result_mng_name] = da

            # Get the total variance of each result
            da_var_tot = da.var('time')
            var_tot[component_mng_name][result_mng_name] = da_var_tot

            # Loop over the sampling frequencies
            high_pass = da.copy(deep=True)
            med.log.info('Filtering {}'.format(component_mng_name))
            filt_freq = {}
            filt[component_mng_name][result_mng_name] = filt_freq
            da_var_filt = xr.DataArray(np.zeros(coords_shape), coords=coords)
            var_filt[component_mng_name][result_mng_name] = da_var_filt
            for freq in freqs:
                med.log.info('- Frequency: {}'.format(freq))
                if freq != 'H':
                    # Low-pass filter
                    low_pass = high_pass.resample(time=freq).mean('time')
                    # Up-sample
                    t = low_pass.indexes['time']
                    if freq == 'Y':
                        low_pass = low_pass.reindex(
                            time=high_pass.indexes['time'],
                            method='bfill').ffill('time')
                    else:
                        new_index = pd.to_datetime({
                            'year': t.year, 'month': t.month, 'day': t.day,
                            'hour': 0.})
                        low_pass = low_pass.reindex(time=new_index)
                        low_pass = low_pass.reindex(
                            time=high_pass.indexes['time'], method='ffill')
                    # High-pass filter
                    high_pass -= low_pass
                else:
                    low_pass = high_pass

                # Store filtered data
                filt_freq[freq] = low_pass

                # Get the variance
                da_var_filt.loc[freq] = low_pass.var('time')

            # Get normalized variance
            da_var_filt_norm = da_var_filt / da_var_tot

            # Write result
            da_var_filt.attrs['units'] = str(da.attrs.get('units')) + '^2'
            da_var_filt_norm.attrs['units'] = ''
            da_var_tot.attrs['units'] = str(da.attrs.get('units')) + '^2'
            result_postfix = result_mng.get_data_postfix(**kwargs)
            # Write spectrum
            filename = 'band_spectrum_{}{}.nc'.format(
                result_mng.name, result_postfix)
            filepath = os.path.join(file_dir, filename)
            med.log.info('Writing {} {} variance and band spectrum to '
                         '{}'.format(component_mng_name, result_mng_name,
                                     filepath))
            xr.Dataset({'variance': da_var_tot,
                        'band_spectrum': da_var_filt,
                        'band_spectrum_normalized': da_var_filt_norm
                        }).to_netcdf(filepath)

    # Print results
    prec = 1
    np.set_printoptions(precision=prec)
    med.log.info('Component\tVariable\tFrequency Range\tVariance (%)')
    for component_mng_name in component_mng_names:
        component_mng = med.component_managers[component_mng_name]
        for result_mng_name in component_mng.result_managers.keys():
            for freq in freqs:
                val = var_filt[component_mng_name][result_mng_name].loc[
                    freq].values * 100
                med.log.info('{}\t{}:'.format(component_mng_name, freq))
                med.log.info(val)

    med.log.info('Component\tVariable\tFrequency Range\t'
                 'Standard Deviation (%)')
    for component_mng_name in component_mng_names:
        component_mng = med.component_managers[component_mng_name]
        for result_mng_name in component_mng.result_managers.keys():
            for freq in freqs:
                val = xr.ufuncs.sqrt(
                    var_filt[component_mng_name][result_mng_name].loc[
                        freq].values) * 100
                med.log.info('{}\t{}:'.format(component_mng_name, freq))
                med.log.info(val)

    return (ds, filt)


def compare_yearly_covariance(med_ref, med_new, component_mng_names=None,
                              **kwargs):
    dim_comp_reg_name = 'component_region'

    # Read hourly capacity factors from ENTSO-E and GSE
    component_mng_names = component_mng_names or set(
        med_ref.component_managers.keys()).intersection(
        med_new.component_managers.keys())
    for component_mng_name in component_mng_names:
        component_name = med_ref.component_managers[
            component_mng_name].component_name
        ref_mean, ref_risk = get_mean_risk(med_ref, component_mng_name)
        new_mean, new_risk = get_mean_risk(med_new, component_mng_name)

        # Compute the probability to find the observed moment
        # in the Gaussian distribution fitted to the moments
        # computed from the climate data
        q = 0.05
        # For the mean
        ref_mean_like = xr.zeros_like(ref_mean)

        n = new_mean.shape[0]

        # Loop over all dimensions
        for it in ref_mean_like.coords['time']:
            for k in ref_mean_like.coords[dim_comp_reg_name]:
                loc = {'time': it, dim_comp_reg_name: k}
                # Fit a normal distribution to the climate moments
                x = new_mean.loc[{dim_comp_reg_name: k}]
                s = x.std(ddof=1)
                ref_mean_like.loc[loc] = - t.ppf(
                    q/2, df=n - 1, scale=s * np.sqrt(1+1/n))

        # For the covariance
        ref_risk_like = xr.zeros_like(ref_risk)
        # Loop over all dimensions
        for it in ref_risk_like.coords['time']:
            for k in ref_risk_like.coords[dim_comp_reg_name]:
                loc = {'time': it, dim_comp_reg_name: k}
                # Fit a normal distribution to the climate moments
                x = new_risk.loc[{dim_comp_reg_name: k}]
                s = x.std(ddof=1)
                ref_risk_like.loc[loc] = -t.ppf(
                    q/2, df=n - 1, scale=s * np.sqrt(1+1/n))

        # Check which risk observations fall inside the confidence interval
        inside = ((ref_risk.mean('time') > (
            new_risk.mean('time') - ref_risk_like.mean('time'))) &
            (ref_risk.mean('time') < (
                new_risk.mean('time') + ref_risk_like.mean('time'))))

        prec = 1
        np.set_printoptions(precision=prec)
        med_ref.log.info('risk for {}'.format(component_name))
        med_ref.log.info('Observed:')
        med_ref.log.info(ref_risk.mean('time').sel(
            component_name=component_name).values * 100)
        med_ref.log.info('Computed:')
        ua = uarray((new_risk.mean('time').sel(component_name=component_name)
                     .values * 100).round(prec), (ref_risk_like.mean(
                         'time').sel(component_name=component_name).values
            * 100).round(prec))
        s = '['
        for (k, (u, ins)) in enumerate(
                zip(ua, inside.sel(component_name=component_name).values)):
            s += repr(u)
            s += ', ' if k < (len(ua) - 1) else ''
        s += ']'
        med_ref.log.info(s)


def get_mean_risk(med, component_mng_name, **kwargs):
    dim_comp_reg_name = 'component_region'
    dim_comp_reg = {dim_comp_reg_name: (
        'component', 'region_multi')}

    # Get result
    comp_ref = med.component_managers[component_mng_name]
    component_name = comp_ref.component_name
    comp_ref.get_result(**kwargs)
    res_name = med.cfg['component_managers'][component_mng_name]
    da = comp_ref.result[res_name]

    # Expand component dimension
    da = da.expand_dims('component').assign_coords(
        component=[component_name])

    # Remove NaNs
    da = da[~da.isnull().any(['region'])]

    # Stack
    da = da.stack(**dim_comp_reg)
    coord_comp_reg = da.coords[dim_comp_reg_name]
    n_comp_reg = len(coord_comp_reg)

    # Sub-sample, if needed
    if med.cfg['frequency'] == 'day':
        da = da.resample(time='D').mean('time')

    # Get mean capacity factors for each year
    gp = da.resample(time='Y')
    da_mean = gp.mean('time')

    # Get the capacity factor covariances for each year
    time = pd.to_datetime(list(gp.groups))
    coords = [('time', time), coord_comp_reg]
    da_risk = xr.DataArray(np.empty((time.shape[0], n_comp_reg)),
                           coords=coords)
    for year, group in gp:
        da_risk.loc[{'time': year}] = group.std('time')

    return da_mean, da_risk



"""Data-source base definitions."""
from abc import ABC, abstractmethod
import os
from collections import MutableMapping, OrderedDict
import functools
import xarray as xr
from .container import Container, ensure_collection


class DataSourceBase(Container, MutableMapping):
    """Base data-source class for APIs."""

    def __init__(self, med, name, cfg=None, variable_names=None,
                 task_names=None, parent=None, **kwargs):
        """Build data source linked to mediator.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :param variable_names: Name(s) of variable(s) composing dataset.
          Default is `None`.
        :param task_names: Names of potential tasks for container to perform.
          Default is `None`.
        :param parent: Parent container.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        :type variable_names: (collection of) :py:class:`str`
        :type task_names: set
        :type parent: :py:class:`.container.Container`
        """
        # Add read tasks per variable
        if variable_names is not None:
            variable_names = ensure_collection(variable_names, set)
            if task_names is None:
                task_names = set()
            for variable_name in variable_names:
                task_names.add('read__{}'.format(variable_name))
                task_names.add('write__{}'.format(variable_name))

        # Initialize as container
        parent = med if parent is None else parent
        super(DataSourceBase, self).__init__(
            med, name, cfg=cfg, task_names=task_names, parent=parent,
            **kwargs)

        #: Whether this is a gridded data source.
        #: By default, data source not gridded.
        self.gridded = False

        #: Data-source data.
        self.data = OrderedDict()

        #: Data-source variable names.
        self.variable_names = set()

        # Add set of variable names composing dataset
        if variable_names is not None:
            self.update_variables(variable_names)

        #: Variable to data-source mapping and variables list
        #: (for it to be useful, we need to assume that one and only one
        #: dataset is associated to a variable.
        #: This may be improved by adding information
        #: on the data source to variables names).
        self.var_data_sources = OrderedDict()

        # Update variable to data-source mapping,
        # for this single data source.
        for variable_name in self.variable_names:
            self.var_data_sources.setdefault(
                variable_name, {}).update({self.name: self})

    def update_variables(self, variable_names, **kwargs):
        """Add variable names to data source.

        :param variable_names: Name(s) of variable(s) to add.
        :type variable_names: (collection of) :py:class:`str`
        """
        variable_names = ensure_collection(variable_names, set)

        # Update variables set
        self.variable_names.update(variable_names)

        # Update tasks
        for variable_name in variable_names:
            task_name = 'read__{}'.format(variable_name)
            if task_name not in self.task_mng:
                self.task_mng.update({task_name: True})
            task_name = 'write__{}'.format(variable_name)
            if task_name not in self.task_mng:
                self.task_mng.update({task_name: True})

    def __getitem__(self, variable_name):
        """Get variable data from :py:attr:`data`.

        :param variable_name: Variable name.
        :type variable_name: str

        :returns: Variable.
        :rtype: :py:class:`xarray.DataArray`

        .. note: Calls `__getitem__` method from :py:class:`dict`.
        """
        return self.data[variable_name]

    def get(self, variable_name, default=None):
        """Get variable data from :py:attr`data`.

        :param variable_name: Variable name.
        :param default: Default value. Default is `None`.
        :type variable_name: str
        :type default: :py:class:`xarray.DataArray`

        :returns: Variable.
        :rtype: :py:class:`xarray.DataArray`

        .. note: Calls `get` method from :py:class:`dict`.
        """
        return self.data.get(variable_name, default)

    def __setitem__(self, variable_name, data):
        """Set item in :py:attr:`data`.

        :param variable_name: Variable name.
        :param data: Data of variable to set.
        :type variable_name: str
        :type data: :py:class:`xarray.DataArray`

        .. note: Calls `__setitem__` method from :py:class:`dict`.
        """
        self.data[variable_name] = data

    def __contains__(self, variable_name):
        """Test if variable in data source.

        :param variable_name: Variable name.
        :type variable_name: str
        """
        return variable_name in self.variable_names

    def __delitem__(self, variable_name):
        """Remove variable from :py:attr:`variables` set
        and from :py:attr:`data` mapping.

        :param variable_name: Variable name.
        :type variable_name: str
        """
        # Remove variable
        self.variable_names.remove(variable_name)
        # Remove data for variable
        del self.data[variable_name]

    def __iter__(self):
        """Iterate :py:attr:`data` mapping."""
        return iter(self.data)

    def __len__(self):
        """Number of variables."""
        return len(self.variable_names)

    def __str__(self):
        """Get dataset as string."""
        s = "<{} '{}'>\n".format(str(self.__class__)[8:-2], self.name)
        s += '{}'.format(self.data)

        return s

    def update(self, var_data):
        """Update data with given dataset.

        :param var_data: Dataset.
        :type var_data: Mapping of :py:class:`xarray.DataArray`
        """
        # Update variable list
        self.variable_names.update(var_data.keys())

        # Update data
        self.data.update(var_data)

    def copy_data(self, data_src, variable_name=None):
        """Copy data from another data source, e.g.
        of type :py:class:`Prediction` or :py:class:`Feature`.

        :param data_src: Data source to copy.
        :param variable_name: Name of a specific variable from which
          to copy data. Default is `None`
        :param data_src: :py:class:`.data_sources.`DataSourceBase`
        :param variable_name: str
        """
        if variable_name is not None:
            self.update({variable_name: data_src.get(variable_name)})
        else:
            self.update(data_src.data)

    def close(self, variable_names=None):
        """Close dataset.

        :param variable_names: Variable(s) to close. Default is `None`,
          in which case all variables are closed.
        :type variable_names: (collection of) :py:class:`str`
        """
        # Get variables list
        variable_names = (ensure_collection(variable_names, set) or
                          self.variable_names)

        # Close each variable separately
        for variable_name in variable_names:
            self.data[variable_name].close()

    def read(self, variable_names=None, **kwargs):
        """Read source dataset as dictionary of :py:class:`xarray.DataArray`.

        :param variable_names: Variable(s) to read. Default is `None`,
          in which case variables are read.
        :type variable_names: (collection of) :py:class:`str`
        """
        variable_names = (ensure_collection(variable_names, set) or
                          self.variable_names)

        if variable_names:
            # Update data source set of variables
            self.update_variables(variable_names)

            # Read each variable separately
            ds = {}
            for iv, variable_name in enumerate(variable_names):
                # Get data filepath
                filepath = '{}.nc'.format(self.get_data_path(
                    variable_name=variable_name, makedirs=False, **kwargs))
                if self.task_mng['read__{}'.format(variable_name)]:
                    if not self.cfg.get('no_verbose'):
                        self.log.info('Reading {} {} from {}'.format(
                            self.name, variable_name, filepath))
                    try:
                        # Try to load as dataarray
                        ds[variable_name] = xr.load_dataarray(filepath)
                    except ValueError:
                        # Or else load as dataset
                        ds[variable_name] = xr.load_dataset(filepath)

                    # Update task manager
                    self.task_mng['read__{}'.format(variable_name)] = False
                else:
                    # Skip
                    if not self.cfg.get('no_verbose'):
                        self.log.info(
                            '{} {} already read: skipping'.format(
                                self.name, variable_name))

            # Update dataset
            self.update(ds)
        else:
            self.log.warning('Empty variable-name collection given: no {} '
                             'data read'.format(self.name))

    def write(self, variable_names=None, **kwargs):
        """Write :py:class:`xarray.DataArray` of each variable in netcdf.

        :param variable_names: Variable(s) to write. Default is `None`,
          in which all variables are written.
        :type variable_names: (collection of) :py:class:`str`

        .. warning:: Exististing files are not overwritten.
          Only existing variables (groups) are.
        """
        variable_names = (ensure_collection(variable_names, set) or
                          self.variable_names)

        if variable_names:
            for iv, variable_name in enumerate(variable_names):
                filepath = '{}.nc'.format(self.get_data_path(
                    variable_name=variable_name, **kwargs))
                if self.task_mng.get('write__{}'.format(variable_name)):
                    if not self.cfg.get('no_verbose'):
                        self.log.info('Writing {} {} to {}'.format(
                            self.name, variable_name, filepath))
                    self.get(variable_name).to_netcdf(filepath, mode='w')

                    # Update task manager
                    self.task_mng['write__{}'.format(variable_name)] = False
                else:
                    # Skip
                    if not self.cfg.get('no_verbose'):
                        self.log.info(
                            '{} {} already written: skipping'.format(
                                self.name, variable_name))
        else:
            self.log.warning('Empty variable-name collection given: no {} '
                             'data written'.format(self.name))

    def get_postfix(self, **kwargs):
        """Return empty postfix string.

        returns: Postfix.
        rtype: str
        """
        return ''

    def get_data_postfix(self, with_src_name=False, **kwargs):
        """Get data-source postfix.
        A user-defined postfix may be defined in the `'postfix'`
        entry of the data source configuration.
        Otherwise, the standard postfix is used by calling
        :py:meth:`get_postfix`.
        The data-source name is prepended.

        :param with_src_name: Whether to prefix postfix with source name.
        :type with_src_name: bool

        :returns: Postfix.
        :rtype: str
        """
        # Get user-defined postfix or standard postfix
        postfix = self.cfg.get('postfix')
        postfix = (self.get_postfix(**kwargs) if postfix is None else
                   postfix)

        # Prepend data-source name
        if with_src_name:
            postfix = '_{}{}'.format(self.name, postfix)

        return postfix

    def get_data_path(self, variable_name=None, makedirs=True, **kwargs):
        """Get data-source filepath.

        :param variable_name: Data variable. Default is `None`.
        :param makedirs: Make directories if needed. Default is `True`.
        :type variable_name: str
        :type makedirs: bool

        :returns: Filepath.
        :rtype: str
        """
        # Variable string
        var_pf = '_{}'.format(variable_name) if variable_name else ''

        # Data-postfix string
        data_pf = self.get_data_postfix(
            variable_name=variable_name, **kwargs)

        # Filepath
        filename = '{}{}{}'.format(self.name, var_pf, data_pf)
        data_dir = self.get_data_dir(makedirs=makedirs, **kwargs)
        filepath = os.path.join(data_dir, filename)

        return filepath


class MultiDataSource(DataSourceBase, MutableMapping):
    def __init__(self, med, data_sources, task_names=set(),
                 default_tasks_value=True, **kwargs):
        """Build data source composed to multiple data sources
        and linked to mediator.

        :param med: Mediator.
        :param data_sources: Data sources dictionary.
        :param task_names: Names of potential tasks for container to perform.
          Default is `set()`.
        :param default_tasks_value: If `True`, ask to perform all tasks.
          Otherwise, none. Default is `True`.
        :type med: :py:class:`.mediator.Mediator`
        :type data_sources: :py:class:`dict` of :py:class:`DataSourceBase`
        :type task_names: set
        :type default_tasks_value: bool

        .. warn:: At the moment, a variable can only belong to one data source.
        """
        # Multiple data-source attributes specification
        cfg = None
        name = self.get_name(data_sources)

        # Initialize as data source
        super(MultiDataSource, self).__init__(
            med, name, cfg=cfg, task_names=task_names,
            default_tasks_value=default_tasks_value, **kwargs)

        #: Data sources composing multiple data source.
        self.data_sources = OrderedDict()

        # Update with data sources
        self.update_data_sources(data_sources)

    @staticmethod
    def get_name(data_sources):
        """Get class name from data_sources.

        :param data_sources: Data-sources list or data-source names list.
        :type data_sources: :py:class:`list` of :py:class:`DataSourceBase`
          or :py:class:`str`

        :returns: Multiple data-source name.
        :rtype: str
        """
        # Note: in principle, data_sources' keys correspond to name
        # attribute of each data_src
        try:
            name = '__'.join(data_src.name
                             for data_src in data_sources.values())
        except AttributeError:
            name = '__'.join(data_sources)

        return name

    def update_data_sources(self, data_sources):
        """Update :py:attr:`data_sources` and :py:attr:`variable_names`.

        :param data_sources: Data sources dictionary.
        :type data_sources: :py:class:`dict` of :py:class:`DataSourceBase`
        """
        for src_name, data_src in data_sources.items():
            # Update data sources
            self.data_sources.update({src_name: data_src})

            # Update variable names
            self.update_variables(data_src.variable_names)

            # Update variable to data-source mapping,
            # allowing for multiple data sources per variable
            for variable_name in data_src.variable_names:
                self.var_data_sources.setdefault(
                    variable_name, {}).update({src_name: data_src})

            # Check if gridded
            if data_src.gridded:
                self.gridded = True

            # Add data sources as children
            self.update_children(self.data_sources)

    def get_data_sources(self, variable_name):
        """Get single data source(s) containing variable.

        :param variable_name: Variable name.
        :type variable_name: str

        :returns: Data source(s) associated with variable.
        :rtype: (mapping of :py:class:`str` to) :py:class:`DataSourceBase`
        """
        data_sources = self.var_data_sources.get(variable_name)

        if len(data_sources) > 1:
            # Multiple data sources for this variable
            sel_data_sources = {src_name: data_src
                                for src_name, data_src in data_sources.items()}
        else:
            # Single data source
            sel_data_sources = list(data_sources.values())[0]

        return sel_data_sources

    def get(self, variable_name, default=None):
        """Get variable from data source containing variable.

        :param variable_name: Variable name.
        :param default: Default (mapping to) array(s). Default is `None`.
        :type variable_name: str
        :type default: (mapping of :py:class:`str` to)
          :py:class:`xarray.DataArray`

        :returns:

          * :py:class:`xarray.DataArray`, if a single data source is
            associated to variable,
          * Mapping from source names to variable-data.

        :rtype: (mapping of :py:class:`str` to) :py:class:`xarray.DataArray`

        .. seealso:: :py:meth:`get_data_sources`
        """
        # Data sources containing variable
        data_sources = self.get_data_sources(variable_name)

        if data_sources is not None:
            if isinstance(data_sources, DataSourceBase):
                # Single data source
                data = data_sources[variable_name]
            else:
                # Multiple data sources for this variable
                data = {src_name: data_src[variable_name]
                        for src_name, data_src in data_sources.items()}
        else:
            data = default

        return data

    def __getitem__(self, variable_name):
        """Get item from data source containing variable.

        :param variable_name: Variable name.
        :type variable_name: str

        :returns:

          * :py:class:`xarray.DataArray`, if a single data source is
            associated to variable,
          * Mapping from source names to variable-data.

        :rtype: (mapping of :py:class:`str` to) :py:class:`xarray.DataArray`
        """
        # Array(s) associated with variable
        data = self.get(variable_name)

        if data is None:
            # Variable not found
            raise KeyError(variable_name)

        return data

    def __iter__(self):
        """Iterate :py:attr:`data` mapping."""
        return iter(self.data_sources)

    def __str__(self):
        """Get dataset as string."""
        s = "<{} '{}'>\n".format(str(self.__class__)[8:-2], self.name)
        s += '\n'.join('{}\n{}'.format(str(data_src), str(data_src.data))
                       for data_src in self.data_sources.values())

        return s

    def __setitem__(self, variable_name, sources_data):
        """Set items in data sources containing variable.

        :param variable_name: Variable name.
        :param sources_data: Source name to variable-data-to-set mapping.
        :type variable_name: str
        :type data: mapping of :py:class:`str` to :py:class:`xarray.DataArray`
        """
        # Data source containing variable
        data_sources = self.var_data_sources[variable_name]

        # Set variable in data source
        for src_name, src_data in sources_data.items():
            data_sources[src_name][variable_name] = src_data

    def update(self, variable_data):
        """Update data with given dataset.

        :param variable_data: Dataset.
        :type variable_data: Mapping of :py:class:`xarray.DataArray`
        """
        for variable_name, sources_data in variable_data.items():
            # Mapping of single data source(s) containing variable
            data_sources = self.variable_data_sources[variable_name]

            # Update single data source
            for src_name, src_data in sources_data.items():
                data_sources[src_name].data.update(
                    {variable_name: src_data})

        # Update multiple data-source variables
        self.update_variables(variable_data.keys())

    def download(self, variable_names=None, **kwargs):
        """Download multiple data source calling :py:meth:`DataSourceBase.download`
        of each data source.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        .. seealso:: :py:meth:`DataSourceBase.download`
        """
        for data_src in self.data_sources.values():
            if isinstance(data_src, DataSourceLoaderBase):
                # Download
                data_src.download(variable_names=variable_names, **kwargs)

    def manage_download(
            self, variable_names=None, **kwargs):
        """Manage multiple data-source download calling
        :py:meth:`DataSourceBase.manage_download` of each data source.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        .. seealso:: :py:meth:`DataSourceBase.manage_download`
        """
        for data_src in self.data_sources.values():
            if isinstance(data_src, DataSourceLoaderBase):
                data_src.manage_download(
                    variable_names=variable_names, **kwargs)

    def load(self, variable_names=None, **kwargs):
        """Load multiple data source calling :py:meth:`DataSourceBase.load`
        of each data source.

        :param variable_names: Names of variables to load.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are loaded.
        :type variable_names: (collection of) :py:class:`str`

        .. seealso:: :py:meth:`DataSourceBase.load`
        """
        d = {}
        for data_src in self.data_sources.values():
            if isinstance(data_src, DataSourceLoaderBase):
                # Load
                d[data_src.name] = data_src.load(
                    variable_names=variable_names, **kwargs)

        return d

    def get_data(self, variable_names=None, **kwargs):
        """Load data from multiple data sources calling
        :py:meth:`DataSourceBase.get_data` of each data source.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        :returns: Dataset :py:attr:`data`.
        :rtype: mapping

        .. seealso:: :py:meth:`DataSourceBase.get_data`
        """
        for data_src in self.data_sources.values():
            if isinstance(data_src, DataSourceLoaderBase):
                # Get data for single source, transmitting keywords
                data_src.get_data(variable_names=variable_names, **kwargs)

        return self.data

    def get_mask(self, **kwargs):
        """Get mask for gridded single data sources calling
        :py:meth:`GriddedDataSourceBase.get_mask`.

        .. seealso:: :py:meth:`GriddedDataSourceBase.get_mask`
        """
        for data_src in self.data_sources.values():
            if data_src.gridded:
                # Get data for single source, transmitting keywords
                data_src.get_mask(**kwargs)

    def read(self, variable_names=None, **kwargs):
        """Read multiple data source.

        .. seealso:: :py:meth:`DataSourceBase.read`
        """
        for src_label, data_src in self.data_sources.items():
            data_src.read(variable_names=variable_names, **kwargs)

    def write(self, variable_names=None, **kwargs):
        """Write multiple data source.

        .. seealso:: :py:meth:`DataSourceBase.write`
        """
        for data_src in self.data_sources.values():
            data_src.write(variable_names=variable_names, **kwargs)

    def get_data_postfix(self, with_src_name=False, **kwargs):
        """Get multiple data-source postfix as sum of each single
        data-source postfix.

        :param with_src_name: Whether to prefix postfix with source name.
        :type with_src_name: bool

        :returns: Postfix.
        :rtype: str
        """
        postfix = ''.join(data_src.get_data_postfix(
            with_src_name=with_src_name, **kwargs)
            for data_src in self.data_sources.values())

        return postfix


class DataSourceLoaderBase(DataSourceBase, ABC):
    """Data-source loader base class. Requires :py:meth:`load` method
    to be implemented. Also includes a passing :py:meth:`download` method."""

    def __init__(self, med, name, cfg=None, variable_names=None, parent=None,
                 task_names=None, **kwargs):
        """Build data source with downloading and loading capabilities.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :param variable_names: List of variables composing dataset.
          Default is `None`.
        :param parent: Parent container.
        :param task_names: Names of potential tasks for container to perform.
          Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        :type variable_names: (collection of) :py:class:`str`
        :type parent: :py:class:`.container.Container`
        :type task_names: set
        """
        # Add download and load tasks per variable
        if variable_names is not None:
            variable_names = ensure_collection(variable_names, set)
            if task_names is None:
                task_names = set()
            for variable_name in variable_names:
                task_names.add('load__{}'.format(variable_name))
                task_names.add('download__{}'.format(variable_name))

        super(DataSourceLoaderBase, self).__init__(
            med, name, cfg=cfg, variable_names=variable_names, parent=parent,
            task_names=task_names, **kwargs)

    def update_variables(self, variable_names, **kwargs):
        """Add variables to data source.

        :param variable_names: (List of) name(s) of variable(s) to add.
        :type variable_names: (collection of) :py:class:`str`
        """
        variable_names = ensure_collection(variable_names, set)

        super(DataSourceLoaderBase, self).update_variables(
            variable_names, **kwargs)

        # Update tasks
        for variable_name in variable_names:
            task_name = 'download__{}'.format(variable_name)
            if task_name not in self.task_mng:
                self.task_mng.update({task_name: True})
            task_name = 'load__{}'.format(variable_name)
            if task_name not in self.task_mng:
                self.task_mng.update({task_name: True})

    @abstractmethod
    def load(self, variable_names=None, **kwargs):
        """Retrieve data from the input source and return an object."""
        raise NotImplementedError

    def download(self, variable_names=None, **kwargs):
        """Download data."""
        self.log.warning('{} download not implemented'.format(self.name))

    def get_data(self, variable_names=None, **kwargs):
        """Read or load data from a given source and store it in
        :py:attr:`data` member.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        :returns: Dataset :py:attr:`data`.
        :rtype: mapping
        """
        # Get list of variables to load and to read
        variable_names = (ensure_collection(variable_names, set) or
                          self.variable_names)
        variable_to_load_names = variable_names.copy()
        for variable_name in variable_names:
            if not self.task_mng.get('load__{}'.format(variable_name)):
                variable_to_load_names.discard(variable_name)
        variable_to_read_names = variable_names.difference(
            variable_to_load_names)

        if variable_to_load_names:
            # Download data if needed
            self.manage_download(
                variable_names=variable_to_load_names, **kwargs)

            if not self.cfg.get('no_verbose'):
                svar = ', '.join(str(variable_name)
                                 for variable_name in variable_to_load_names)
                self.log.info('Loading {} {}'.format(self.name, svar))

            # Load data
            data = self.load(
                variable_names=variable_to_load_names, **kwargs)
            self.update(data)

            # Write data (all components, in case more than one)
            self.write(variable_names=variable_to_load_names, **kwargs)

            # Update task manager with all loaded variables
            # (even the ones that were not requested)
            for variable_name in data:
                self.task_mng['load__{}'.format(variable_name)] = False

        if variable_to_read_names:
            # Read variables
            if not self.cfg.get('no_verbose'):
                svar = ', '.join(str(variable_name)
                                 for variable_name in variable_to_read_names)
                self.log.info(
                    '{} {} already loaded'.format(self.name, svar))
            self.read(variable_names=variable_to_read_names, **kwargs)

        return self.data

    def manage_download(self, variable_names=None, **kwargs):
        """Manage data download.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`
        """
        # Get list of variables to download
        variable_names = (ensure_collection(variable_names, set) or
                          self.variable_names)
        variable_to_download_names = variable_names.copy()
        for variable_name in variable_names:
            if not self.task_mng.get('download__{}'.format(variable_name)):
                variable_to_download_names.discard(variable_name)

        # Download variables
        if variable_to_download_names:
            if not self.cfg.get('no_verbose'):
                svar = ', '.join(
                    str(variable_name) for variable_name in
                    variable_to_download_names)
                self.log.info('Downloading {} {}'.format(
                    ', '.join(self.name.split('__')), svar))

            # Download
            downloaded_variable_names = self.download(
                variable_names=variable_to_download_names, **kwargs)

            # Update task manager with all downloaded variables
            # (even the ones that were not requested)
            if downloaded_variable_names is not None:
                for variable_name in downloaded_variable_names:
                    self.task_mng['download__{}'.format(variable_name)] = False


class Composer(object):
    """Compose functions.
    `Compose(f, g, h, **kwargs)(ds) = h(g(f(ds, **kwargs), **kwargs), **kwargs)`.
    """

    def __init__(self, *args, **kwargs):
        """Constructor."""
        #: Functions to compose.
        self._functions = args

        #: Composed function.
        self.composed = functools.reduce(
            lambda f, g: lambda ds=None, **kwargs_add: g(
                ds=f(ds=ds, **kwargs, **kwargs_add), **kwargs, **kwargs_add),
            self._functions, lambda ds=None, **kwargs_add: ds)

    def __call__(self, ds=None, **kwargs_add):
        """Caller."""
        return self.composed(ds=ds, **kwargs_add)



"""Component and result managers base definitions."""
from collections import MutableMapping, OrderedDict
from .container import Container, ensure_collection, load_config
from .data_source import DataSourceBase, MultiDataSource
from .actuator_base import add_actuator


class ResultManager(DataSourceBase):
    """Manager for prediction results or output variables associated with a
    component."""

    def __init__(self, component_mng, name, cfg=None, **kwargs):
        """Constructor.

        :param component_mng: Component manager to which variable is
          associated.
        :param name: Result manager name.
        :param cfg: Result manager configuration. Default is `None`.
        :type component_mng: :py:class:`ComponentManager`
        :type name: str
        :type cfg: mapping
        """
        #: Component manager of result manager
        self.component_mng = component_mng

        #: Actuators.
        self.actuators = OrderedDict()

        #: Data sources.
        self.data_sources = OrderedDict()

        #: Features for each stage. Used in feature-extraction cases.
        self.feature = OrderedDict()

        #: Stage to extractor mapping.
        self.extractor = OrderedDict()

        #: Prediction result data source.
        self.prediction = None

        #: Result data source. Either :py:attr:`prediction`
        #:  or :py:attr:`feature['output']`.
        self.result = None

        # Initialize as container with component manager as parent
        super(ResultManager, self).__init__(
            med=self.component_mng.med, name=name, cfg=cfg,
            parent=component_mng, **kwargs)

        #: Result name.
        self.result_name = self.cfg.get('result_name') or self.name

        # Initialize actuators from configuration
        for actuator_name in self.cfg:
            actuator = add_actuator(
                self, actuator_name=actuator_name, **kwargs)

            # Add given actuator as attribute
            setattr(self, actuator_name, actuator)

            # Reference actuator in actuators set
            self.actuators[actuator_name] = actuator

        # Add data sources required by result manager (from all of its
        # actuators)
        for actuator in self.actuators.values():
            if actuator.data_sources is not None:
                for data_src in actuator.data_sources.values():
                    # Add as single data source
                    self.data_sources.update({data_src.name: data_src})

                    # Add as single data sources from multiple data source
                    if isinstance(data_src, MultiDataSource):
                        self.data_sources.update({
                            single_data_src.name: single_data_src
                            for single_data_src in
                            data_src.data_sources.values()})

        # Initialize extractors
        self._init_extractors(**kwargs)

        # Initialize data attributes
        self._init_data_attributes(**kwargs)

    def _init_extractors(self, **kwargs):
        """Initialize extractors."""
        # Add output extractor
        self.extractor['output'] = self.output_extractor

        if hasattr(self, 'feature_extractor'):
            # Add feature extractors for all stages
            for stage in 'fit', 'predict':
                self.extractor[stage] = self.feature_extractor

    def _init_data_attributes(self, **kwargs):
        """Initialize data attributes.

        .. warning:: :py:meth:`_init_extractors` must be called before.
        """
        # Initialize output-feature data source
        self.feature['output'] = Feature(self, stage='output', **kwargs)

        if hasattr(self, 'estimator'):
            # Initialize prediction data source
            self.prediction = Prediction(self, **kwargs)

            # Assign prediction to result
            self.result = self.prediction
        else:
            # Assign output feature to result
            self.result = self.feature['output']

        # Plug result methods to output variable
        self.update_variables(self.result.variable_names, **kwargs)
        self.data = self.result.data
        self.get_data_postfix = self.result.get_data_postfix
        self.get_data_path = self.result.get_data_path

        # Initialize features, associated with inputs
        if hasattr(self, 'feature_extractor'):
            for stage in self.feature_extractor.stages:
                self.feature[stage] = Feature(self, stage=stage, **kwargs)

    def extract(self, stage, **kwargs):
        """Extract or read feature from input data,
        store it in the :py:attr:`feature` member, and save it to file.

        :param stage: Modeling stage: `'fit'`, `'predict'` or `'output'`.
        :type stage: str
        """
        extractor = self.extractor[stage]
        if extractor.task_mng.get('extract__' + stage):
            # Extract feature
            # Get input data source
            data_src = extractor.data_sources[stage]
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Extracting {} feature to {} {} from {}'.format(
                        self.component_mng.name, stage, self.name,
                        data_src.name))

            # Extract feature from the input data
            self.feature[stage].update(
                extractor.transform(data_src=data_src, stage=stage, **kwargs))

            # Write all variables of feature
            if extractor.task_mng.get('write__' + stage):
                self.feature[stage].write(**kwargs)

            # Update task manager
            extractor.task_mng['extract__' + stage] = False
        else:
            # Read feature for variable
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    '{} feature to {} {} already extracted'.format(
                        self.component_mng.name, stage, self.name))
            self.feature[stage].read(**kwargs)

    def fit(self, **kwargs):
        """Fit estimator.

        .. note:: Input and output data for the fitting are read from file.
          They should thus be extracted before.
        """
        # Read or fit
        stage = 'fit'
        if hasattr(self, 'estimator'):
            if self.estimator.task_mng.get(stage):
                # Get input feature
                self.extract(stage, **kwargs)

                # Get output data
                self.extract('output', **kwargs)

                # Fit estimator
                if not self.cfg.get('no_verbose'):
                    self.log.info('Fitting {} {} estimator'.format(
                        self.component_mng.name, self.name))
                self.estimator.fit(self.feature[stage], self.feature['output'],
                                   **kwargs)

                # Save the fitted estimator
                self.estimator.write(**kwargs)

                # Update task managern
                self.estimator.task_mng[stage] = False
            else:
                # Load the estimator coefficients from file
                if not self.cfg.get('no_verbose'):
                    self.log.info('{} {} estimator already fitted'.format(
                        self.component_mng.name, self.name))
                self.estimator.read(**kwargs)
        else:
            if not self.cfg.get('no_verbose'):
                self.log.info('No estimator provided: skipping.')

    def predict(self, variable_names=None, **kwargs):
        """Predict and store result to :py:attr:`prediction` member.

        :param variable_names: Variable names. Default is `None`.
        :type variable_names: (collection of) :py:class:`str`

        .. note:: Input data for the prediction is read from file.
          It should thus be extracted before.
        """
        stage = 'predict'
        if hasattr(self, 'estimator'):
            if self.estimator.task_mng.get(stage):
                # Get estimator
                self.fit(**kwargs)

                # Get feature to predict
                self.extract(stage, **kwargs)

                # Apply
                if not self.cfg.get('no_verbose'):
                    self.log.info('Predicting {} {}'.format(
                        self.component_mng.name, self.name))
                self.prediction.update(self.estimator.predict(
                    self.feature[stage], **kwargs))

                # Save
                # Warning: input data to fit estimator will be forgotten
                self.prediction.write(variable_names=self.name, **kwargs)

                # Update task manager
                self.estimator.task_mng[stage] = False
            else:
                # Read prediction
                if not self.cfg.get('no_verbose'):
                    self.log.info('{} {} already predicted'.format(
                        self.component_mng.name, self.name))
                self.prediction.read(variable_names=self.name, **kwargs)
        else:
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'No {} {} estimator provided: skipping.'.format(
                        self.component_mng.name, self.name))

    def get_data(self, variable_names=None, **kwargs):
        """Get output from prediction if :py:attr:`estimator` is not `None`,
        or directly from (extracted) output data source, and store it in
        :py:attr:`result`.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        :returns: Dataset :py:attr:`data`.
        :rtype: mapping
        """
        if hasattr(self, 'estimator'):
            # Predict
            self.predict(**kwargs)
        else:
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'No {} {} estimator provided: directly extracting '
                    'output data'.format(self.component_mng.name, self.name))
            # Get data
            self.extract('output', **kwargs)

        return self.data


class ComponentManager(Container, MutableMapping):
    """Component manager."""

    def __init__(self, med, name, cfg=None, result_mng_names=None,
                 area=None, **kwargs):
        """Initialize component manager attached to mediator.

        :param med: Mediator.
        :param name: Component-manager name.
        :param cfg: Component-manager configuration.
          Default is `None`, in which case it is loaded by
          :py:func:`.container.load_config`.
        :param result_mng_names: Result-manager name(s)
          to estimate for component.
        :param area: Area to which the component manager is assigned.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        :type result_mng_names: (collection of) :py:class:`str`
        :type area: str
        """
        # Load component-manager configuration
        loaded_cfg = (cfg or load_config(med, name))

        #: Result managers of component manager.
        self.result_managers = OrderedDict()

        #: Component-manager data sources.
        #: Union of data sources from all result managers.
        self.data_sources = OrderedDict()

        # Initialize as container with mediator as parent
        super(ComponentManager, self).__init__(
            med, name, cfg=loaded_cfg, parent=med, **kwargs)

        #: Component name.
        self.component_name = self.cfg.get('component_name') or self.name

        #: Covered area.
        self.area = area

        #: Places in area.
        self.place_names = None
        if area and getattr(self.med, 'geo_src'):
            self.place_names = list(self.med.geo_src.area_places_sources[area])

        # Add result managers
        result_mng_names = ensure_collection(result_mng_names)
        for result_mng_name in result_mng_names:
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Injecting {} result manager in {} component '
                    'manager'.format(result_mng_name, self.name))
            self.result_managers[result_mng_name] = ResultManager(
                self, result_mng_name, cfg=self.cfg[result_mng_name])

        # Add data sources required by component manager (from all of its
        # result managers)
        for result_mng in self.result_managers.values():
            self.data_sources.update(result_mng.data_sources)

    def __getitem__(self, result_mng_name):
        """Get result manager from :py:attr:`result_managers`.

        :param result_mng_name: Result-manager name.
        :type result_mng_name: str

        :returns: Result manager.
        :rtype: :py:class:`ResultManager`
        """
        return self.result_managers[result_mng_name]

    def get(self, result_mng_name, default=None):
        """Get result manager from :py:attr`data`.

        :param result_mng_name: Result-manager name.
        :param default: Default value. Default is `None`.
        :type result_mng_name: str
        :type default: :py:class:`ResultManager`

        :returns: Result manager.
        :rtype: :py:class:`ResultManager`
        """
        return self.result_managers.get(result_mng_name, default)

    def __setitem__(self, result_mng_name, result_mng):
        """Set result manager in :py:attr:`data`.

        :param result_mng_name: Result-manager name.
        :param result_mng: Result manager to set.
        :type result_mng_name: str
        :type result_mng: :py:class:`ResultManager`
        """
        self.result_managers[result_mng_name] = result_mng

    def __contains__(self, result_mng_name):
        """Test if result manager in result managers.

        :param result_mng_name: Result-manager name.
        :type result_mng_name: str
        """
        return result_mng_name in self.result_managers

    def __delitem__(self, result_mng_name):
        """Remove result manager from :py:attr:`result_managers`.

        :param result_mng_name: Result-manager name.
        :type result_mng_name: str
        """
        del self.result_managers[result_mng_name]

    def __iter__(self):
        """Iterate :py:attr:`result_managers` mapping."""
        return iter(self.result_managers)

    def __len__(self):
        """Number of result managers."""
        return len(self.result_managers)


class ResultDataSourceBase(DataSourceBase):
    """Base result data."""

    def __init__(self, result_mng, name=None, variable_names=None, **kwargs):
        """Initialize feature as data source.

        :param result_mng: Result manager to which data is associated.
        :param name: Data name.
        :param variable_names: Name(s) of variable(s) composing dataset.
          Default is `None`.
        :type result_mng: :py:class:`.component.ResultManager`
        :type name: str
        :type variable_names: (collection of) :py:class:`str`
        """
        # Attach component manager and result manager
        #: Data-source result manager.
        self.result_mng = result_mng
        #: Component manager.
        self.component_mng = self.result_mng.component_mng

        kwargs.update({
            'med': self.component_mng.med, 'name': name,
            'variable_names': variable_names, 'cfg': self.result_mng.cfg,
            'task_names': None})
        super(ResultDataSourceBase, self).__init__(**kwargs)

    def get_data_dir(self, makedirs=True, **kwargs):
        """Get path to data directory.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Data directory path.
        :rtype: str
        """
        return self.med.cfg.get_project_data_directory(
            self.component_mng, makedirs=makedirs)


class Feature(ResultDataSourceBase):
    """Feature data source."""

    def __init__(self, result_mng, stage, **kwargs):
        """Initialize feature as data source.

        :param result_mng: Result manager to which data is associated.
        :param stage: Modeling stage: `'fit'` or `'predict'`.
        :type result_mng: str
        :type stage: str
        """
        #: Data-source estimation-stage.
        self.stage = stage

        #: Extractor.
        self.extractor = result_mng.extractor[stage]

        # Variables from extractors or as result_mng
        variable_names = (self.extractor.variable_names or
                          set([result_mng.result_name]))

        # Build result-manager data source with extractor variables
        name = '{}_{}'.format(result_mng.component_mng.name, self.stage)
        super(Feature, self).__init__(
            result_mng=result_mng, name=name,
            task_names=None, variable_names=variable_names, **kwargs)

        # Plug get_data_path method from extractor if possible
        if hasattr(self.extractor, 'get_data_path'):
            self.get_data_path = self.extractor.get_data_path

    def get_data_postfix(self, **kwargs):
        """Get feature postfix (overwrite :py:class:`DataSourceBase`
        implementation).

        :returns: Postfix.
        :rtype: str
        """
        # Get input data postfix
        data_src = self.extractor.data_sources[self.stage]
        feature_postfix = data_src.get_data_postfix(
            with_src_name=True, **kwargs)

        # Add modifier postfix
        if (hasattr(self.result_mng, 'modifier')
                and (self.stage in ['predict', 'output'])):
            feature_postfix += (self.result_mng.modifier.
                                get_extractor_postfix(**kwargs))

        # Add feature postfix
        feature_postfix += self.extractor.get_extractor_postfix(**kwargs)

        return feature_postfix

    def get_data_path(self, *args, **kwargs):
        """Get data path from base or from original data source if no
        extraction is performed."""
        if self.extractor.no_extraction:
            return self.extractor.data_sources[self.stage].get_data_path(
                *args, external=True, **kwargs)
        else:
            return super(Feature, self).get_data_path(*args, **kwargs)


class Prediction(ResultDataSourceBase):
    """Prediction data source."""

    def __init__(self, result_mng, **kwargs):
        """Initialize prediction as data source.

        :param result_mng: Result manager to which data is associated.
        :type result_mng: str
        """
        # Build result-manager data source
        name = '{}_prediction'.format(result_mng.component_mng.name)
        super(Prediction, self).__init__(
            result_mng=result_mng, name=name,
            variable_names=result_mng.result_name, task_names=None, **kwargs)

    def get_data_postfix(self, **kwargs):
        """Get prediction postfix.

        :returns: Postfix.
        :rtype: str
        """
        return self.result_mng.estimator.get_fit_postfix(**kwargs)


class DataSourceWithComponentsMixin():
    """Add components management to data source loader."""

    def get_component_names_per_variable(
            self, variable_names=None, component_names=None, **kwargs):
        """Get set of component names from argument or from all mediator
        components that have :py:obj:`self` as data source and sort them
        by variable.

        :param variable_names: Variable names.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are returned.
        :param component_names: Component names.
          Default is `None`, in which case all mediator components
          that have :py:obj:`self` as data source are returned.
        :type variable_names: (collection of) :py:class:`str`
        :type component_names: (collection of) :py:class:`str`

        :returns: Component names.
        :type component_names: collection of :py:class:`str`
        """
        # Ensure collection for given/data-source variable names
        variable_names = (ensure_collection(variable_names, set)
                          or self.variable_names)

        # Get variable-component names mapping
        return self.med.get_component_names_per_variable_for_source(
            self.name, variable_names, component_names, **kwargs)

    def finalize_array(self, da, variable_name, **kwargs):
        """Finalize array adding/converting units if possible,
          transposing, naming. Used by APIs.

        :param da: Data array.
        :param variable_name: Variable name.
        :type da: :py:class:`xarray.DataArray`
        :type variable_name: str
        """
        # Convert units if possible
        try:
            da = da * float(self.cfg['unit_conversions'][variable_name])
        except (KeyError, TypeError):
            pass

        # Add units if possible
        try:
            da.attrs['units'] = str(self.cfg['units'][variable_name])
        except KeyError:
            da.attrs['units'] = 'Unknown'

        # Transpose
        da = da.transpose('time', 'component', 'region')

        # Name array as variable
        da.name = variable_name

        return da


def finalize_dataset(func):
    """"Decorator to finalize arrays in dataset returned by
    :py:meth:`..data_source.DataSourceLoaderBase.load`.

    :param func: Function to be decorated.
    :type func: function

    .. seealso:: :py:meth:`finalize_array`
    """

    # Define decorating function
    def inner(self, variable_component_names=None, **kwargs):
        # Call function to get dataset
        ds = func(self, variable_component_names, **kwargs)

        # Finalize arrays for all variables
        for variable_name, da in ds.items():
            ds[variable_name] = self.finalize_array(
                da, variable_name, **kwargs)

        return ds

    return inner


def parse_variable_component_args(func):
    """"Decorator to parse variable_names and/or variable_component_names
    arguments for :py:meth:`..data_source.DataSourceLoaderBase.download`
    and :py:meth:`..data_source.DataSourceLoaderBase.load`.

    :param func: Function to be decorated.
    :type func: function
    """

    # Define decorating function
    def inner(self, variable_names=None, component_names=None,
              variable_component_names=None, **kwargs):
        if variable_component_names is None:
            # Get variable-component names mapping
            variable_component_names = self.get_component_names_per_variable(
                variable_names, component_names, **kwargs)
        else:
            # Ensure component names as collections
            for variable_name, component_names in (
                    variable_component_names.items()):
                variable_component_names[
                    variable_name] = ensure_collection(component_names, set)

        # Call function and return
        return func(self, variable_component_names, **kwargs)

    return inner


def download_to_compute_capacity_factor(func):
    """Decorator to handle capacity factor from capacity and generation.
    in :py:meth:`..data_source.DataSourceLoaderBase.download`
    and :py:meth:`..data_source.DataSourceLoaderBase.load`.

    :param func: Function to be decorated.
    :type func: function
    """

    # Define decorating function
    def inner(self, variable_component_names=None, **kwargs):
        # Add capacity factor to variable to load names
        variable_component_to_load_names = add_capacity_factor(
            variable_component_names, **kwargs)

        # Call function and return
        return func(self, variable_component_to_load_names, **kwargs)

    return inner


def compute_capacity_factor(func):
    """Decorator to handle capacity factor from capacity and generation.
    in :py:meth:`..data_source.DataSourceLoaderBase.download`
    and :py:meth:`..data_source.DataSourceLoaderBase.load`.

    :param func: Function to be decorated.
    :type func: function
    """

    # Define decorating function
    def inner(self, variable_component_names=None, **kwargs):
        # Add capacity factor to variable to load names
        variable_component_to_load_names = add_capacity_factor(
            variable_component_names, **kwargs)

        # Call function
        ds = func(self, variable_component_to_load_names, **kwargs)

        # Get capacity factor
        variable_name = 'capacity_factor'
        if variable_name in variable_component_names:
            # Get capacity factor
            da = self._get_capacity_factor(ds, **kwargs)

            # Add variable to dataset
            ds[variable_name] = self.finalize_array(
                da, variable_name, **kwargs)

        # Select variables
        ds_sel = {}
        for variable_name in variable_component_names:
            ds_sel[variable_name] = ds[variable_name]

        return ds_sel

    return inner


def add_capacity_factor(variable_component_names, **kwargs):
    """Add capacity factor to :py:obj:`variable_component_to_load_names`.

    :param variable_component_names: Names of components to load per
      variable.
    :type variable_component_names: mapping from :py:class:`str`
      to collection

    :returns: Names of components to load per variable with
      capacity factor added.
    :rtype: mapping from :py:class:`str` to collection
    """
    # Add capacity and generation to variables to load
    variable_name = 'capacity_factor'
    variable_component_to_load_names = variable_component_names.copy()
    if variable_name in variable_component_names:
        # Make sure that generation and capacity are loaded
        variable_component_to_load_names.update(
            {'generation': variable_component_names[variable_name],
             'capacity': variable_component_names[variable_name]})

        # Remove capacity factor from variables to load
        del variable_component_to_load_names['capacity_factor']

    return variable_component_to_load_names



"""Estimator of capacity factor time series from generation time series
and capacity factor averages."""
import numpy as np
import pandas as pd
from ..actuator_base import ExtractorBase
from ..utils.learn import select_data


class Actuator(ExtractorBase):
    """Estimator of capacity factor time series from generation time series
    and capacity factor averages."""
    #: Default result name.
    DEFAULT_RESULT_NAME = 'capacity_factor'

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        if result_mng.result_name != self.DEFAULT_RESULT_NAME:
            self.log.warning(
                'Result name {} given to constructor does not correspond '
                'to {} to be estimated by {}'.format(
                    result_mng.result_name, self.DEFAULT_RESULT_NAME,
                    self.name))

        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            variable_names={result_mng.result_name, 'capacity'}, **kwargs)

        #: Required input-variables.
        self.input_variable_names = set(['generation', 'capacity_factor'])

    def transform(self, data_src, **kwargs):
        """Extract capacity factor time series from generation time series
        and capacity factor averages for component.

        :param data_src: Input data source containing generation
          time series and capacity factor averages.
        :type data_src: :py:class:`.data_source.DataSourceBase`

        :returns: Capacity factor and capacity time series.
        :rtype: :py:class:`dict` from :py:class:`str` to
          :py:class:`xarray.DataArray`
        """
        if not self.cfg.get('no_verbose'):
            self.log.info(
                'Computing {} {} time series from {} generation time series '
                'and {} capacity factor means'.format(
                    self.component_mng.name, self.result_mng.name,
                    *list(data_src.var_data_sources['generation']),
                    *list(data_src.var_data_sources['capacity_factor'])))

        # Get data (letting multiple data source manage variables)
        kwargs_data_src = kwargs.copy()
        kwargs_data_src.pop('variable_names', None)
        data_src.get_data(**kwargs_data_src)

        # Select component and result and reorder regions and coordinates
        da_gen = select_data(data_src, self.result_mng,
                             variable_name='generation', **kwargs)
        da_cf_mean = select_data(data_src, self.result_mng,
                                 variable_name='capacity_factor', **kwargs)

        # Get mean generation
        da_gen_mean = da_gen.resample(
            time=self.cfg['mean_frequency']).mean('time')

        # Get mean capacity factor, thus making sure that
        # to down-sample if needed and to use the same
        # time-index format as for the generation
        da_cf_mean = da_cf_mean.resample(
            time=self.cfg['mean_frequency']).mean('time')

        # Get mean capacity
        da_cap_mean = da_gen_mean / da_cf_mean

        # # Get capacity time series by linearly interpolating
        # da_cap = da_cap_mean.resample(time='H').interpolate('linear')

        t_cap_mean = da_cap_mean.indexes['time']
        start = '{}-01-01'.format(t_cap_mean[0].year)
        # # Get capacity time series from last year to avoid 2015
        # # high generation
        # start = '{}-01-01'.format(t_cap_mean[0].year + 1)
        end = t_cap_mean[-1] + pd.Timedelta(23, unit='H')
        t_sub = pd.date_range(start=start, end=end, freq='H')
        da_cap = da_cap_mean.reindex(time=t_sub).bfill('time').ffill('time')

        # Get capacity factors time series on intersection
        da_cf_sub = da_gen / da_cap

        if self.cfg.get('extrapolate'):
            da_cf = _extrapolate(da_cf_mean, da_cf_sub, **kwargs)
        else:
            da_cf = da_cf_sub

        # Format
        da_cf.name = self.result_mng.result_name
        da_cf.attrs['units'] = ''
        da_cap_mean.name = 'capacity'

        return {da_cf.name: da_cf, da_cap_mean.name: da_cap_mean}

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        postfix = '{}_{}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs),
            self.med.cfg['frequency'])

        if self.cfg.get('extrapolate'):
            postfix += '_extrapolate'

        return postfix


def _extrapolate(da_mean, da_sub, time_name='time', **kwargs):
    """Extrapolate hourly data to years for which not hourly
    data is available by multiplying the mean hourly profile over
    a year by the yearly-mean data for which more years are available.

    :param da_mean: Time-averaged data array.
    :param da_sub: Data array over non-extrapolated period.
    :param time_name: Time-dimension name. Default is `time`.
    :type da_mean: :py:class:`xarray.DataArray`
    :type da_sub: :py:class:`xarray.DataArray`
    :type time_name: str

    :returns: Extrapolated data array.
    :rtype: :py:class:`xarray.DataArray`

    .. warning:: Time index should start with new year.
    """
    t_mean = da_mean.indexes[time_name]
    start = '{}-01-01'.format(t_mean[0].year)
    end = t_mean[-1] + pd.Timedelta(23, unit='H')
    t_ext = pd.date_range(start=start, end=end, freq='H')
    t_sub = da_sub.indexes[time_name]
    da = da_sub.reindex(**{time_name: t_ext})
    da = _filter_leap_years(da, time_name=time_name)

    # Get year to use to extrapolate
    da_cycle = da.sel(**{time_name: str(t_sub[0].year)})

    # Compute correctors for each year based on yearly mean
    corrector = da_mean / da_sub.mean(time_name)

    # Apply correction to missing years
    gp = da.resample(**{time_name: 'Y'})
    for date, da_date in gp:
        # Time index of year
        date = pd.Timestamp(date)
        t_date = da_date.indexes[time_name]

        # Add year only if absent
        if date.year not in t_sub.year:
            # Correct year
            da_corr = da_cycle * corrector.sel(**{time_name: date}, drop=True)

            # Add year
            da_corr[time_name] = (time_name, t_date)
            da.loc[{time_name: str(date.year)}] = da_corr

    return da


def _filter_leap_years(da, time_name='time', **kwargs):
    """Remove additional days from leap years.

    :param da: Data array from which to remove days.
    :param time_name: Time-dimension name. Default is `time`.
    :type da: :py:class:`xarray.DataArray`
    :type time_name: str

    :returns: Filtered data array.
    :rtype: :py:class:`xarray.DataArray`
    """
    t = da.indexes[time_name]
    leap_years = np.unique(t.year[t.is_leap_year])
    da_filt = da.copy(deep=True)
    for year in leap_years:
        sy = '{}-02-29'.format(year)
        da_filt = da_filt[(da_filt.indexes[time_name].date !=
                           pd.Timestamp(sy).date())]

    return da_filt



"""Statistical learning."""
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.metrics import r2_score


class ShiftingBiasCorrector():
    """Bias corrector via shifting in order for the input and output
    means to coincide.
    """

    def __init__(self):
        """Constructor."""
        #: Intercept set when fitting.
        self.coef_ = None

    def fit(self, X, y):
        """Fit.

        :param X: Input data.
        :param y: output data.
        :type X: array_like
        :type y: array_like

        :returns: :py:obj:`self`.
        :rtype: :py:class:`ShiftingBiasCorrector`

        .. warning:: :py:obj:`X` and :py:obj:`y` should have the same
          dimensions.
        """
        self.coef_ = y.mean() - X.mean()

        return self

    def predict(self, X, return_std=False):
        """Predict multi-output variable using a model
            trained for each target variable.

        :param X: Input data.
        :type X: array_like

        :returns: Prediction.
        :rtype: array_like
        """
        return X + self.coef_

    def score(self, X, y):
        """r2 score.

        :param X: Input data.
        :param y: output data.
        :type X: array_like
        :type y: array_like

        :returns: r2 score.
        :rtype: float
        """
        return r2_score(y, self.predict(X))

    def get_params(self, deep=True):
        return {}

    def set_params(self, d):
        return self


class MultiInputRegressor():
    """Multi input estimator.

    :param estimator: An estimator implementing `fit` and `predict`.
    :type estimator: estimator object
    """

    def __init__(self, estimator):
        """Constructor.

        :param estimator: Estimator.
        :type estimator: :py:class:`sklearn.base.BaseEstimator`
        """
        #: Base estimator.
        self.estimator = estimator
        #: Multiple estimators.
        self.estimators = None
        #: Scores for each estimator.
        self.scores = None

    def fit(self, X, y):
        """Fit model to data.
        Fit a separate model for each input and output variables.

        :param X: Input array,
            shape(n_samples, n_feature, n_outputs).
        :param y: Output array, shape(n_samples, n_outputs)
        :type X: array_like
        :type y: array_like

        :returns: :py:obj:`self`
        :rtype: :py:class:`MultiInputRegressor`
        """
        # Loop over inputs and outputs
        self.estimators = []
        for ie, (X_l_T, y_l_T) in enumerate(zip(X.T, y.T)):
            # Clone estimator
            e = clone(self.estimator)

            # Fit data
            e.fit(X_l_T.T, y_l_T.T)

            # Save estimator
            self.estimators.append(e)

        # Save number of estimators
        self.n_estimators = len(self.estimators)

        return self

    def predict(self, X, return_std=False):
        """Predict multi-output variable using a model
            trained for each target variable.

        :param X: Input data, shape(n_samples, n_feature, n_outputs).
        :param return_std: If `True`, return standard deviation
           of posterior prediction(in the Bayesian case).
           Default is `False`.
        :type X: array_like
        :type return_std: bool

        :returns: Prediction.
        :rtype: array_like
        """
        y_pred = np.empty((X.shape[0], X.shape[2]))
        if return_std:
            y_std = np.empty((X.shape[0], X.shape[2]))
        for ie, X_l_T in enumerate(X.T):
            # Manage return std for Bayesian case only
            kwargs = {'return_std': True} if return_std else {}

            # Predict
            pred = self.estimators[ie].predict(X_l_T.T, **kwargs)

            # Manage prediction result
            if return_std:
                y_pred[:, ie] = pred[0]
                y_std[:, ie] = pred[1]
            else:
                y_pred[:, ie] = pred
        if return_std:
            return (y_pred, y_std)
        else:
            return y_pred

    def score(self, X, y):
        """Compute individual scores and variance-weighted score
        of multiple-output prediction.

        :param X: Input data, shape(n_samples, n_feature, n_outputs).
        :param y: True target data, shape(n_samples, n_outputs).
        :type X: array_like
        :type y: array_like

        :returns: Variance-weighted score.
        :rtype: float

        .. note:: Individual scores are saved in :py:attr:`scores`.
        """
        # Predict
        y_pred = self.predict(X)

        # Save raw scores
        self.scores = r2_score(y, y_pred, multioutput='raw_values')

        # Return variance weighted score
        score = (self.scores * y.var(0)).sum() / y.var(0).sum()

        return score


def last_full_years_date(time, first_date, **kwargs):
    """Get last date in date-time index which is a multiple of a year
    away from start date.

    :param time: Date-time index.
    :param first_date: First date.
    :type time: :py:class:`pandas.DatetimeIndex`
    :type first_date: :py:class:`datetime.date`

    :returns: Last date.
    :rtype: :py:class:`datetime.date`
    """
    last_date = pd.date_range(time[0], time[-1], freq='A-DEC')[-1]

    # Make sure last samples are present
    freq = time.inferred_freq
    if freq is None:
        if (time[1] - time[0]).seconds == 3600:
            freq = 'H'
    if freq == 'H':
        last_date += pd.Timedelta('23H')

    return last_date


def intersection_time_slice(da_in, da_out):
    """Get full years intersection between two arrays, if possible.

    :param da_in: First array.
    :param da_out: Second array.
    :type da_in: :py:class:`xarray.DataArray`
    :type da_out: :py:class:`xarray.DataArray`

    :returns: Intersection time slice.
    :rtype: slice

    .. warning:: The time index of the input and output arrays
      must be compatible. For instance, daily-sampled indices
      but with different hours won't intersect.
    """
    # Match frequencies
    t_out = da_out.indexes['time']
    freq_out = t_out.inferred_freq
    t_in = da_in.indexes['time']
    freq_in = t_in.inferred_freq
    if freq_out in ['A-DEC', 'Y']:
        # Make index of yearly mean start 1st of January
        da_out['time'] = pd.DatetimeIndex(
            ['{}-01-01'.format(t.year) for t in t_out])
        if freq_in == 'H':
            start = '{}-01-01 00:00:00'.format(t_out[0].year)
            end = '{}-12-31 23:00:00'.format(t_out[-1].year)
            t_out_hour = pd.date_range(start, end, freq=freq_in)
            da_out = da_out.reindex(time=t_out_hour, method='ffill')

    t_inter = da_out.indexes['time'].intersection(
        da_in.indexes['time'])
    if len(t_inter) > 0:
        first_date_inter = t_inter[0]
        last_date_inter = last_full_years_date(
            t_inter, first_date_inter)
        if ((last_date_inter - first_date_inter)
                >= pd.Timedelta(1, unit='Y')):
            # If there are at least a year of common data,
            # use common data only
            time_slice = slice(first_date_inter, last_date_inter)

            return time_slice


def parse_fit_data(
        data_src_in, data_src_out, result_mng, subsample_freq=None,
        select_period=False, time_slice=None, **kwargs):
    """Get input and output data arrays.

    :param data_src_in: Training set input data source.
    :param data_src_out: Training set output data source.
    :param result_mng: Result manager.
    :param subsample_freq: Sub-sampling frequency. Default is `None`, in which
      case no sub-sampling is performed.
    :param time_slice: Time slice to select period. Default is `None` in which
      case the full period is kept.
    :param select_period: Whether to select period. Default is `False`.
    :type data_src_in: :py:class:`..data_source.DataSourceBase`
    :type data_src_out: :py:class:`..data_source.DataSourceBase`
    :type result_mng: :py:class:`..component.ResultManager`
    :type subsample_freq: str
    :type select_period: bool
    :type time_slice: slice

    :returns: Input and output data arrays.
    :rtype: :py:class:`tuple` containing two
      :py:class:`xarray.DataArray`'s
    """
    # Select result, copy and ensure dimensions order
    da_in = select_data(data_src_in, result_mng, **kwargs)
    da_out = select_data(data_src_out, result_mng, **kwargs)

    # Subsample if requested
    if subsample_freq:
        da_in = da_in.resample(time=subsample_freq).mean('time')
        da_out = da_out.resample(time=subsample_freq).mean('time')

    if select_period:
        # Select input and output periods
        # as given by configuration or first full years
        da_in, time_slice = select_period_from_array(
            da_in, time_slice=time_slice, **kwargs)
        da_out, _ = select_period_from_array(
            da_out, time_slice=time_slice, **kwargs)

    return da_in, da_out


def select_period_from_array(da, time_slice=None, **kwargs):
    """Select full years period from an array.
    Try to get first and last dates from configuration.
    Otherwise, select the first full years.

    :param da: Array.
    :param time_slice: Time slice to select period. Default is `None` in which
      case the full period is kept.
    :type da: :py:class:`xarray.DataArray`
    :type time_slice: :py:class:`slice` or collection

    :returns: Tuple containing the selected array and time slice.
    :rtype: :py:class:`tuple` of :py:class:`xarray.DataArray`
      and :py:class:`slice`
    """
    if time_slice is None:
        # Get time index
        time = da.indexes['time']

        # First date is given or first available date
        first_date = pd.Timestamp(time[0])

        # Last date is given or last date which is a multiple of one
        # year past first date
        last_date = pd.Timestamp(
            last_full_years_date(time, first_date, **kwargs))
    else:
        try:
            first_date, last_date = time_slice.start, time_slice.stop
        except AttributeError:
            first_date, last_date = time_slice

    # Slice
    time_slice = slice(first_date, last_date)
    da = da.sel(time=time_slice)

    return da, time_slice


def select_data(data_src, result_mng, variable_name=None, **kwargs):
    """Select variable from data source, copy it, reorder regions,
    drop components and transpose, if possible.

    :param data_src: Training set input data source.
    :param result_mng: Result manager.
    :param variable_name: Variable to select from data source.
      Default is `None`, in which case `result_mng.result_name`
      is used instead.
    :type data_src: :py:class:`..data_source.DataSourceBase`
    :type result_mng: :py:class:`..component.ResultManager`
    :type variable_name: str

    :returns: Input and output data arrays.
    :rtype: :py:class:`tuple` containing two
      :py:class:`xarray.DataArray`'s
    """
    component_mng = result_mng.component_mng

    # Select variable or result
    variable_name = variable_name or result_mng.result_name
    da = data_src[variable_name].copy(deep=True)

    # Try to reorder regions from component-manager place names
    try:
        da = da.sel(region=component_mng.place_names)
    except ValueError:
        pass

    # Try to select component in case needed
    try:
        da = da.sel(component=component_mng.component_name, drop=True)
    except ValueError:
        pass
    try:
        da = da.drop('component')
    except ValueError:
        pass

    # Try to transpose
    try:
        da = da.transpose('time', 'region')
    except ValueError:
        pass

    return da



"""natomraden.se API for Sweden's bidding zones."""
from pkg_resources import resource_filename
import geopandas as gpd
from ..geo import DefaultMaskAPI, DefaultMaskMaker, GEO_VARIABLE_NAME


class DataSource(DefaultMaskMaker, DefaultMaskAPI):
    #: Default source name.
    DEFAULT_SRC_NAME = 'natomraden'

    #: Area: `'Sweden'`.
    AREA = 'Sweden'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor calling :py:class:`DefaultMaskMaker`'s
        constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    def download(self, *args, **kwargs):
        """Convenience function to warn that natomraden.se data is not downloaded
        but loaded from package data.

        :returns: Names of downloaded variables.
        :rtype: :py:class:`set` of :py:class:`str`

        .. note:: The JSON geometries were downloaded from the `natomraden.se <https://www.natomraden.se/>` source code.
        """
        self.log.warning(
            '{} is not to be downloaded. It is instead directly loaded from '
            'the package data.'.format(self.DEFAULT_SRC_NAME))

        return {GEO_VARIABLE_NAME}

    def read_file(self, *args, **kwargs):
        """Read geographical data for area from package resources.

        :param area: Geographical area for which to read the data.
        :type area: str

        :returns: Geographic data frame.
        :rtype: :py:class:`geopandas.GeoDataFrame`
        """
        # Get resource name
        resource_name = self.get_download_resource_name(self, **kwargs)

        # Read geographical file for area
        gdf = gpd.read_file(resource_filename(__name__, resource_name))

        return gdf

    def get_download_resource_name(self, variable_name, **kwargs):
        """Get package resource name for data scratched from natomraden.se.

        :returns: Resource name.
        :rtype: str
        """
        filename = 'sweden_bidding_zones_geometry.json'
        resource_name = '../data/{}/{}/{}'.format(
            self.AREA, self.DEFAULT_SRC_NAME, filename)

        return resource_name



"""SMARD API for Germany, Luxembourg and Austria."""
import os
import requests
import zipfile
from io import BytesIO
import numpy as np
import pandas as pd
import xarray as xr
from ..data_source import DataSourceLoaderBase
from ..component import (
    DataSourceWithComponentsMixin, parse_variable_component_args,
    download_to_compute_capacity_factor, compute_capacity_factor,
    finalize_dataset)


class DataSource(DataSourceLoaderBase, DataSourceWithComponentsMixin):
    """SMARD data source."""
    #: Default source name.
    DEFAULT_SRC_NAME = 'smard'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

    @parse_variable_component_args
    @download_to_compute_capacity_factor
    def download(self, variable_component_names, **kwargs):
        """Download data.

        :param variable_component_names: Names of components to download per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to :py:class:`set` of :py:class:`str`

        :returns: Names of downloaded components per variable.
        :rtype: mapping from :py:class:`str` to :py:class:`set` of
          :py:class:`str`

        .. note:: The data is downloaded from
          `SMARD.de <>https://www.smard.de/en/downloadcenter/download_market_data/>`.

        """
        # Get place names for source
        _, src_region_place = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, **kwargs))

        # Get years
        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)

        # Loop over variables
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_download_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are downloaded
                component_to_download_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to download
                component_to_download_names.discard(component_name)

            # Loop over components
            for component_name in component_to_download_names:
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {}'.format(component_name))

                # Loop over dates
                for y, year in enumerate(years):
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- {}'.format(year))

                    # Loop over regions
                    for region_name in src_region_place:
                        # Get URL
                        url, json = self._get_url_json(
                            variable_name, component_name, year, region_name,
                            **kwargs)
                        filepath = self._get_download_filepath(
                            variable_name, component_name, year, region_name,
                            **kwargs)
                        if not self.cfg.get('no_verbose'):
                            self.log.info('---- from {} to {}'.format(
                                url, filepath))

                        # Request and raise exception if needed
                        response = requests.post(url, json=json)
                        response.raise_for_status()

                        # Unzip single file
                        zip_ref = zipfile.ZipFile(BytesIO(response.content))
                        zip_ref.filelist[0].filename = os.path.basename(
                            filepath)
                        zip_ref.extract(zip_ref.filelist[0],
                                        os.path.dirname(filepath))
                        zip_ref.close()

        # Return names of downloaded variables
        return variable_component_names

    @parse_variable_component_args
    @compute_capacity_factor
    @finalize_dataset
    def load(self, variable_component_names, **kwargs):
        """Load data.

        :param variable_component_names: Names of components to load per
          variable.
        :type variable_component_names: mapping from :py:class:`str`
          to collection

        :returns: Time series for each variable and component.
        :rtype: dict
        """
        # Get place names for source
        src_place_regions, src_region_place = (
            self.med.geo_src.get_place_regions_for_source(
                self.name, **kwargs))

        # Get years
        years = range(int(self.cfg['first_year']),
                      int(self.cfg['last_year']) + 1)

        # Loop over variables
        ds = {}
        for variable_name, component_names in variable_component_names.items():
            if not self.cfg.get('no_verbose'):
                self.log.info('- {}:'.format(variable_name))

            # Manage wind component as onshore plus offshore
            component_name = 'wind'
            component_to_load_names = component_names.copy()
            if component_name in component_names:
                # Make sure that generation and capacity are loaded
                component_to_load_names.update(
                    ['wind-onshore', 'wind-offshore'])

                # Remove capacity factor from components to load
                component_to_load_names.discard(component_name)

            # Loop over components
            da = None
            for component_name in component_to_load_names:
                if not self.cfg.get('no_verbose'):
                    self.log.info('-- {}'.format(component_name))

                df_comp = None
                for y, year in enumerate(years):
                    if not self.cfg.get('no_verbose'):
                        self.log.info('--- {}'.format(year))

                    # Loop over regions
                    df_year = pd.DataFrame(columns=src_region_place.values(),
                                           dtype=float)
                    for region_name, place_name in src_region_place.items():
                        # Get filepath
                        filepath = self._get_download_filepath(
                            variable_name, component_name, year, region_name,
                            **kwargs)
                        if not self.cfg.get('no_verbose'):
                            self.log.info('---- from {}'.format(filepath))

                        # Read data frame and select series
                        df_region = pd.read_csv(
                            filepath, **self.cfg['read_csv_kwargs'])
                        s = df_region.iloc[:, 0]

                        # Add region data to place
                        if df_year[place_name].any():
                            df_year[place_name] += s
                        else:
                            df_year[place_name] = s

                    # Add year
                    df_comp = (df_year if df_comp is None else
                               pd.concat([df_comp, df_year], 'index'))

                if variable_name == 'generation':
                    # Convert time zone to UTC
                    df_comp.index = df_comp.index.tz_localize(
                        self.cfg['timezone'],
                        ambiguous='infer').tz_convert(None)
                df_comp.index.name = 'time'

                # Convert to data array
                da_comp = xr.DataArray(
                    df_comp, dims=('time', 'region')).expand_dims(
                    'component').assign_coords(component=[component_name])

                # Add component to data array
                da = (da_comp if da is None else
                      xr.concat([da, da_comp], dim='component'))

            if variable_name == 'generation':
                # Convert 15m-power to 1h-energy
                da = da.resample(time='H').sum('time')

            # Add total wind component if needed
            component_name = 'wind'
            if component_name in component_names:
                # Treat NaNs as zero unless in both components
                da_on = da.sel(component='wind-onshore', drop=True)
                da_off = da.sel(component='wind-offshore', drop=True)
                na = da_on.isnull() & da_off.isnull()
                da_wind = (da_on.fillna(0.) + da_off.fillna(0.)).where(
                    ~na, np.nan)
                da_wind = da_wind.expand_dims('component').assign_coords(
                    component=[component_name])
                da = xr.concat([da, da_wind], dim='component')

            # Select requested components
            da = da.sel(component=list(component_names))

            # Add to dataset
            ds[variable_name] = da

        return ds

    def _get_capacity_factor(self, ds, **kwargs):
        """Get capacity factor from generation and capacity.

        :param ds: Dataset containing generation and capacity.
        :type ds: mapping

        :returns: Capacity factor.
        :rtype: :py:class:`xarray.DataArray`
        """
        da_gen = ds['generation']
        da_cap = ds['capacity']

        # Upsample capacity to generation using capacities
        # from the 1st of January of next year
        freq = da_gen.indexes['time'].inferred_freq
        da_cap_up = da_cap.copy()
        da_cap_up['time'] = [pd.Timestamp('{}-01-01T00:00'.format(t.year - 1))
                             for t in da_cap_up.indexes['time']]
        da_cap_up = da_cap_up.resample(time=freq).ffill().reindex(
            time=da_gen.indexes['time'], method='ffill')

        # Get number of hours in generation time step
        nhours = 1.

        # Get capacity factor
        da = da_gen / (da_cap_up * nhours)

        return da

    def _get_url_json(self, variable_name, component_name, year, region_name,
                      **kwargs):
        """Get URL and JSON parameters for HTTP POST request.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :param year: Year.
        :param region_name: Region name.
        :type variable_name: str
        :type component_name: str
        :type year: int
        :type region_name: str

        :returns: URL and JSON parameters.
        :rtype: :py:class:`tuple` with a :py:class:`str` and a mapping
        """
        # Define URL
        url = self.cfg['url']

        # Get from and to timestamps (POSIX time in milliseconds)
        timestamp_from = pd.Timestamp('{}-01-01'.format(
            year)).timestamp() * 1000
        timestamp_to = pd.Timestamp('{}-01-01'.format(
            year + 1)).timestamp() * 1000 - 1.

        # Define JSON parameters
        json = {"request_form": [{
            "format": self.cfg['file_format'].upper(),
            "language": self.cfg['language'].lower(),
            "moduleIds": [
                self.cfg['component_ids'][variable_name][component_name]],
            "region": region_name,
            "timestamp_from": timestamp_from, "timestamp_to": timestamp_to,
            "type":"discrete"}]}

        return url, json

    def _get_download_filepath(
            self, variable_name, component_name, year, region_name, **kwargs):
        """Get download filepath.

        :param variable_name: Name of variable to query.
        :param component_name: Name of component to query.
        :param year: Year.
        :param region_name: Region name.
        :type variable_name: str
        :type component_name: str
        :type year: int
        :type region_name: str

        :returns: Filepath.
        :rtype: str

        .. note:: Dates in filepath are in UTC.
        """
        # Define filename
        filename = '{}_{}_{}_{}_{}.csv'.format(
            self.DEFAULT_SRC_NAME, variable_name, component_name, region_name,
            year)

        # Get filepath
        src_dir = self.med.cfg.get_external_data_directory(self)
        filepath = os.path.join(src_dir, filename)

        return filepath

    def get_postfix(self, **kwargs):
        """Get data-source standard postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}-{}'.format(self.cfg['first_year'],
                                  self.cfg['last_year'])

        return postfix



"""Angle computations for solar applications."""
import numpy as np
import xarray as xr
import xarray.ufuncs as xrf


class SolarTracker(object):
    """Solar tracking computer. Computes the tracking surface-tilt,
    surface-azimuth and angle-of-incidence cosine. When getting an attribute,
    the variable corresponding to it is automatically computed.
    New computations are only performed if the needed attribute value is
    `None`, i.e. if it has not previously been computed.
    To perform new computations, start from a new object,
    or call :py:meth:`clear`."""

    def __init__(self, solar_computer, tilt=None, azimuth=None,
                 tracking='none', angles_in_degrees=False, **kwargs):
        """Initialize solar tracking computer.

        :param solar_computer: Solar angles computer.
        :param tilt: Surface tilt. For a tilt equal to the latitude,
          set to `'lat'`. Should be provided if :py:obj:`tracking` is
          `'vertical'` or `'none'`. Default is `None`.
        :param azimuth: Surface azimuth.
          Should be provided if :py:obj:`tracking` is `'none'`.
          Default is `None`.
        :param tracking: For east-west axis, north-south
          axis, vertical axis, or two axes, respectively set to `'ew_track'`,
          `'ns_track'`, `'vertical'`, or `'two'`. Default is `none`.
        :param angles_in_degrees: Whether angles are given in degrees instead
          of radians.
        :type solar_computer: :py:class:`.solar_calendar.SolarCalendarComputer`
        :type tilt: (sequence of) :py:class:`float`, :py:class:`str`
        :type azimuth: (sequence of) :py:class:`float`, :py:class:`str`
        :type tracking: str
        :type angles_in_degrees: bool
        """
        # Solar computer
        self.sc = solar_computer

        # Clear variables
        self._surface_tilt = None
        self._surface_azimuth = None
        self._cos_aoi = None

        # Manage tracking strategy
        self.tracking = tracking
        fact = np.deg2rad(1.) if angles_in_degrees else 1.
        if tilt == 'lat':
            self._surface_tilt = self.sc.lat
        if tracking == 'vertical':
            if self._surface_tilt is None:
                # In case not previously set by 'lat' case
                self._surface_tilt = tilt * fact
        elif 'none':
            if self._surface_tilt is None:
                # In case not previously set by 'lat' case
                self._surface_tilt = tilt * fact
            self._surface_azimuth = azimuth * fact

    @property
    def surface_tilt(self):
        """Get surface tilt.
        Computation only performed if requested attribute is `None`.
        """
        if self._surface_tilt is None:
            self._surface_tilt = getattr(
                self, self.tracking + '_track_tilt')()

        return self._surface_tilt / self.sc.fact

    @property
    def surface_azimuth(self):
        """Get surface azimuth.
        Computation only performed if requested attribute is `None`.
        """
        if self._surface_azimuth is None:
            self._surface_azimuth = getattr(
                self, self.tracking + '_track_azimuth')()

        return self._surface_azimuth / self.sc.fact

    @property
    def cos_aoi(self):
        """Get angle-of-incidence cosine.
        Computation only performed if requested attribute is `None`.
        """
        if self._cos_aoi is None:
            self._cos_aoi = getattr(
                self, self.tracking + '_track_cos_aoi')()

        return self._cos_aoi

    def ns_track_cos_aoi(self):
        """Get angle-of-incidence cosine for north-south tracking."""
        return xrf.sqrt(1. - xrf.cos(self.sc.declination)**2
                        * xrf.sin(self.sc.hour_angle)**2)

    def ew_track_cos_aoi(self):
        """Get angle-of-incidence cosine for east-west tracking."""
        return xrf.sqrt(
            xrf.cos(self.sc.zenith)**2
            + xrf.cos(self.sc.declination)**2 * xrf.sin(self.sc.hour_angle)**2)

    def vertical_track_cos_aoi(self):
        """Get angle-of-incidence cosine for east-west tracking."""
        return xrf.sqrt(
            xrf.cos(self.sc.zenith) * xrf.cos(self._surface_tilt)
            + xrf.sin(self.sc.zenith) * xrf.sin(self._surface_tilt))

    def two_track_cos_aoi(self):
        """Get angle-of-incidence cosine for east-west tracking."""
        return xr.ones_like(self.sc.zenith)

    def none_track_cos_aoi(self):
        """Get cosine of Angle of Incidence (AOI) of direct beams
        on an array."""
        return (
            xrf.sin(self.sc.zenith) * xrf.sin(self._surface_tilt)
            * xrf.cos(self._surface_azimuth - self.sc.azimuth)
            + xrf.cos(self.sc.zenith) * xrf.cos(self._surface_tilt))

    def ns_track_tilt(self):
        """Get horizontal north-south tracking/east-west axis tilt.

        :returns: Surface tilt.
        :rtype: (sequence of) :py:class:`float`
        """
        return xrf.atan(
            xrf.tan(self.sc.zenith) * xrf.fabs(xrf.cos(self.sc.azimuth)))

    def ns_track_azimuth(self):
        """Get horizontal north-south tracking/east-west axis azimuth.

        :returns: Surface azimuth.
        :rtype: (sequence of) :py:class:`float`
        """
        return (0. if xrf.fabs(self.sc.azimuth) < np.pi / 2 else np.pi)

    def ew_track_tilt(self):
        """Get horizontal east-west tracking/north-south axis tilt.

        :returns: Surface tilt.
        :rtype: (sequence of) :py:class:`float`
        """
        return xrf.atan(
            xrf.tan(self.sc.zenith)
            * xrf.fabs(xrf.cos(self.ew_track_azimuth() - self.sc.azimuth)))

    def ew_track_azimuth(self):
        """Get horizontal east-west tracking/north-south axis azimuth.

        :returns: Surface azimuth.
        :rtype: (sequence of) :py:class:`float`
        """
        return (1. if self.sc.azimuth > 0 else -1.) * np.pi / 2

    def vertical_track_tilt(self):
        """Get vertical-axes tracking tilt.

        :returns: Surface tilt.
        :rtype: (sequence of) :py:class:`float`
        """
        return self._surface_tilt

    def vertical_track_azimuth(self):
        """Get vertical-axes tracking azimuth.

        :returns: Surface azimuth.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.sc.azimuth

    def two_track_tilt(self):
        """Get two-axes tracking tilt.

        :returns: Surface tilt.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.sc.zenith

    def two_track_azimuth(self):
        """Get two-axes tracking azimuth.

        :returns: Surface azimuth.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.sc.azimuth

    def none_track_tilt(self):
        """No-tracking tilt.

        :returns: Surface tilt.
        :rtype: (sequence of) :py:class:`float`
        """
        return self._surface_tilt

    def none_track_azimuth(self):
        """No-tracking azimuth.

        :returns: Surface azimuth.
        :rtype: (sequence of) :py:class:`float`
        """
        return self._surface_azimuth



"""Modifier to regrid."""
from ..actuator_base import ExtractorBase


class Actuator(ExtractorBase):

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg, **kwargs)

    def transform(self, ds=None, **kwargs):
        """Apply modifier to data source.

        :param ds: Dataset to transform.
        :type ds: :py:class:`xarray.Dataset`

        :returns: Modified dataset.
        :rtype: Same as :py:obj:`data_src.data`
        """
        if self.cfg['method'] == 'coarsen':
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Coarsening {} {} features'.format(
                        self.component_mng.name, self.result_mng.name))
            ds_new = ds.coarsen(**self.cfg['kwargs']).mean()
        elif self.cfg['method'] == 'interp':
            if not self.cfg.get('no_verbose'):
                self.log.info(
                    'Interpolating {} {} features'.format(
                        self.component_mng.name, self.result_mng.name))
            ds_new = ds.interp(**self.cfg['kwargs'])

        return ds_new

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        skwargs = ''.join(['_{}{}'.format(k, v)
                           for k, v in self.cfg['kwargs'].items()])
        return '{}_{}{}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs),
            self.cfg['method'], skwargs)



"""Modifier offsetting variable definition."""
from copy import deepcopy
from ..actuator_base import ExtractorBase


class Actuator(ExtractorBase):
    """Modifier offsetting variable"""

    def __init__(self, result_mng, name, cfg=None, **kwargs):
        """Naming constructor.

        :param result_mng: Result manager of variable to estimate.
        :param name: Actuator name.
        :param cfg: Actuator configuration. Default is `None`.
        :type result_mng: :py:class:`ResultManager`
        :type name: str
        :type cfg: mapping
        """
        super(Actuator, self).__init__(
            result_mng=result_mng, name=name, cfg=cfg,
            **kwargs)

        #: Variable to modify.
        self.modified_variable = self.cfg['variable_name']

        #: Offset to add to the variable.
        self.offset = self.cfg['offset']

    def transform(self, data_src=None, ds=None, **kwargs):
        """Apply modifier to data source.

        :param data_src: Data source to modify. Default is `None`,
          in which case :py:obj:`ds` should be given.
        :param ds: Dataset. Default is `None`,
          in which case :py:obj:`data_src` should be given.
        :type data_src: :py:class:`..data_source.DataSource`
        :type ds: :py:class:`xarray.Dataset`

        :returns: Modified dataset.
        :rtype: Same as :py:obj:`data_src.data`
        """
        if not self.cfg.get('no_verbose'):
            self.log.info('Adding {} to {}'.format(
                self.offset, self.modified_variable))

        # Copy source dataset (all variables)
        if ds is None:
            try:
                # As xarray.Dataset
                ds = data_src.data.copy(deep=True)
            except TypeError:
                # As dictionary
                ds = deepcopy(data_src.data)

        # Modify variable in dataset
        ds[self.modified_variable] += self.offset

        return ds

    def get_extractor_postfix(self, **kwargs):
        """Get extractor postfix.

        returns: Postfix.
        rtype: str
        """
        return '{}_offset_{}_{:d}'.format(
            super(Actuator, self).get_extractor_postfix(**kwargs),
            self.modified_variable, int(self.offset * 10 + 0.1))



# @package .
#  e4clim: Energy for CLimate Integrated Model.
#
#  e4clim: Energy for CLimate Integrated Model.

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""Container and task manager classes definitions."""
import os
import oyaml as yaml
from collections import MutableMapping, Iterable, OrderedDict


class Container(object):
    """Container including a mediator, a configuration,
    and a name."""

    def __init__(self, med, name, cfg=None, task_names=set(),
                 default_tasks_value=True, children=None,
                 parent=None, **kwargs):
        """Initialize mediator, configuration and name.

        :param med: Mediator.
        :param name: Container name.
        :param cfg: Container configuration. Default is `None`.
        :param task_names: Names of potential tasks for container to perform.
          Default is `set()`.
        :param default_tasks_value: If `True`, ask to perform all tasks.
          Otherwise, none. Default is `True`.
        :param children: Set of all containers attached to this container.
        :param parent: Parent container.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        :type task_names: set
        :type default_tasks_value: bool
        :type children: :py:class:`set` of :py:class:`Container`
        :type parent: :py:class:`Container`
        """
        #: Mediator.
        self.med = med

        #: Mediator logger.
        self.log = self.med.log

        #: Configuration.
        self.cfg = cfg

        #: Name.
        self.name = name

        #: Containers depending this one.
        self.children = OrderedDict()

        #: Parent container.
        self.parent = None

        # Set parent
        self.set_parent(parent)

        # Update children
        self.update_children(children)

        #: Task manager.
        self.task_mng = TaskManager(
            self, task_names, default_tasks_value=default_tasks_value)

    def set_parent(self, parent, **kwargs):
        if parent is not None:
            # Set parent
            self.parent = parent

            # Add child to parent
            self.parent.update_children({self.name: self})

    def update_children(self, children, **kwargs):
        """Add children to data source and set parent of each child.

        :param children: Children containers to add.
        :type variable_name: mapping of :py:class:`Container`
        """
        if children is not None:
            # Add children
            self.children.update(children)

            # Set parent
            for child in children.values():
                child.parent = self

    def reload_config(self, cfg=None):
        """Reload container configuration and children configurations
        from given/existing mediator configuration.

        :param cfg: Mediator configuration.
        :type cfg: mapping

        .. seealso:: :py:func:`load_config`
        """
        # Reload container configuration
        self.cfg = load_config(self.med, self.name, self.parent)

        # Reload children configurations
        for child in self.children.values():
            child.cfg = load_config(self.med, child.name, self)

    def get_data_dir(self, makedirs=True, **kwargs):
        """Get path to data directory.

        :param makedirs: Make directories if needed. Default is `True`.
        :type makedirs: bool

        :returns: Data directory path.
        :rtype: str
        """
        return self.med.cfg.get_project_data_directory(
            self, makedirs=makedirs, **kwargs)


class TaskManager(MutableMapping):
    """Task manager."""

    def __init__(self, container, task_names, default_tasks_value=True,
                 **kwargs):
        """Initialize task manager and ask to perform all tasks, or not.

        :param container: Container.
        :param task_names: Task names.
        :param default_tasks_value: If `True`, ask to perform all tasks.
          Otherwise, none. Default is `True`.
        :type container: :py:class:`Container`
        :type task_names: set
        :type default_tasks_value: bool
        """
        # Attach container to its task manager
        self.container = container
        self.name = '{}__task_manager'.format(self.container.name)

        # Initialize tasks
        self.task_names = (ensure_collection(task_names, set)
                           if task_names is not None else set())
        self.task = OrderedDict()
        if self.task_names:
            self.set_all(default_tasks_value)

    def update(self, task_dict):
        """Set given task(s) to given value(s). If a task name
        does not exist yet, it is created and its state set to value.

        :param task_dict: Task (name, value) pairs.
        :type task_dict: :py:class`dict` with :py:class:`bool` values
        """
        for task_name, value in task_dict.items():
            # Add each single task
            self[task_name] = value

    def __setitem__(self, task_name, value):
        """Set single given task.

        :param task_name: Task name to set.
        :param value: Task state value.
        :type task_name: str
        :type value: bool
        """
        # Add task if needed
        if task_name not in self.task_names:
            self.task_names.add(task_name)

        # Set task to value
        self.task[task_name] = value

    def set_all(self, value=True, task_names=None):
        """Set all tasks to given value.

        :param value: State value with which to set all tasks.
          Default is `True` for all tasks.
        :param task_names: Name of task(s) to set. Default is `None`,
          in which case all tasks are set.
        :type value: bool
        :type task_names: (collection of) :py:class:`str`

        .. note:: A task is made more precise by adding `__`
          separated words, e.g. `'load__capacity_factor'`.
          To update all `'load'` tasks, `'load'` can then simply be given.
        """
        # Interesect given and available task names
        valid_task_names = (self.task_names if task_names is None else
                            ensure_collection(task_names, set))
        avail_task_names = set()
        valid_task_names_list = [task.split('__') for task in valid_task_names]
        task_names_list = [task.split('__') for task in self.task_names]
        for task_name in valid_task_names_list:
            for avail_task in task_names_list:
                if avail_task[:len(task_name)] == task_name:
                    avail_task_names.add('__'.join(avail_task))

        # Update container tasks
        self.update({tn: value for tn in avail_task_names})

        # Update children tasks
        for child in self.container.children.values():
            child.task_mng.set_all(value, task_names)

    def __getitem__(self, key):
        """Get item from :py:attr:`task`."""
        return self.task[key]

    def __contains__(self, key):
        """:py:attr:`task` contains."""
        return key in self.task

    def __delitem__(self, key):
        # Remove task from dictionary
        del self.task[key]
        # Remove task name from set
        self.task_names.discard(key)

    def __iter__(self):
        """Iterate :py:attr:`task`."""
        return iter(self.task)

    def __len__(self, key, value):
        """Length of :py:attr:`task`."""
        return len(self.task)

    def __str__(self):
        """Get string of container and its children's tasks."""
        s = '{}: {}'.format(self.container.name, str(self.task))
        for child in self.container.children.values():
            s += '\n  ' + '\n  '.join(str(child.task_mng).split('\n'))

        return s

    def get(self, key, default=None):
        """Get item from :py:attr:`task`."""
        return self.task.get(key, default)


def ensure_collection(arg, collection_type=list):
    """Make a list containing argument or cast iterable to list.

    :param arg: Argument from which to make list.
    :param collection_type: Collection type. Default is `list`.
    :type arg: object
    :type collection_type: type

    :returns: Collection containing from argument(s).
    :rtype: :py:obj:`collection_type`

    .. note:: If `None` is given, `None` is returned.
    """
    if arg is not None:
        if isinstance(arg, str) or (not isinstance(arg, Iterable)):
            arg = collection_type([arg])
        else:
            arg = collection_type(arg)

    return arg


def load_config(med, name, parent=None):
    """Load container configuration from main configuration dictionary,
    from given parent, or from file.

    :param med: Mediator.
    :param name: Container name.
    :param parent: Parent container from the configuration of which,
      the configuration of this child container could be gotten.
      Default is `None`.
    :type med: :py:class:`.mediator.Mediator`
    :type name: str
    :type parent: :py:class:`Container`

    :returns: Container configuration.
    :rtype: dict
    """
    cfg_from_parent = parent.cfg.get(name) if parent is not None else None
    user_cfg_path = med.cfg['cfg_path'].get(name)
    if name in med.cfg:
        # Get source configuration from main configuration
        cont_cfg = med.cfg[name]
    elif user_cfg_path:
        # Read source configuration from path
        user_cfg_path = os.path.join(*ensure_collection(user_cfg_path))
        with open(user_cfg_path, 'r') as f:
            cont_cfg = yaml.load(f, Loader=yaml.FullLoader)
    elif parent is not None:
        # Get source configuration from component configuration
        cont_cfg = cfg_from_parent
    else:
        # None worked, no configuration found
        cont_cfg = None
        med.log.warning('No {} configuration found'.format(name))

    return cont_cfg



"""CORDEX API."""
import os
from collections import OrderedDict
import ftplib
import numpy as np
import pandas as pd
import xarray as xr
import iris
from ..grid import GriddedDataSourceBase, get_geodetic_crs, get_geodetic_array
from ..container import ensure_collection


class DataSource(GriddedDataSourceBase):
    """CORDEX data source.

        .. seealso::
            * `cordex variables requirements table <https://www.medcordex.eu/cordex_variables_requirement_table_110628.pdf>`
            * `cordex archive specifications <http://cordex.dmi.dk/joomla/images/cordex/cordex_archive_specifications.pdf>`
    """
    #: Default source name.
    DEFAULT_SRC_NAME = 'cordex'

    def __init__(self, med, name=None, cfg=None, **kwargs):
        """Naming constructor.

        :param med: Mediator.
        :param name: Data source name.
        :param cfg: Data source configuration. Default is `None`.
        :type med: :py:class:`.mediator.Mediator`
        :type name: str
        :type cfg: mapping
        """
        name = name or self.DEFAULT_SRC_NAME
        super(DataSource, self).__init__(med, name, cfg=cfg, **kwargs)

        # Set geographic dimensions of regional model
        self.dims = self.cfg['dims'][self.cfg['rcm_model_name']]

        #: Start and end dates.
        if 'rcp' in self.cfg['cmip5_experiment_name']:
            self.start_dates = self.cfg['streams']['rcp']['start_dates']
            self.end_dates = self.cfg['streams']['rcp']['end_dates']
        else:
            self.start_dates = self.cfg['streams'][self.cfg[
                'cmip5_experiment_name']]['start_dates']
            self.end_dates = self.cfg['streams'][self.cfg[
                'cmip5_experiment_name']]['end_dates']

    def download(self, variable_names=None, **kwargs):
        """Download cordex data and save it locally.

        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are downloaded.
        :type variable_names: (collection of) :py:class:`str`

        :returns: Names of downloaded variables.
        :rtype: :py:class:`set` of :py:class:`str`

        .. note:: The Med-cordex data is available on the
          `Med-cordex website <https://www.medcordex.eu/>`.
        """
        # Get variable names
        variable_names = ensure_collection(
            variable_names, set) or self.variable_names

        # Define data directories
        sim_dir = self._get_sim_dir()

        # Get credentials
        cred = self.med.cfg.get_credentials(
            self.DEFAULT_SRC_NAME, keys=['user', 'passwd'])

        # Login to the FTP server
        ftp = ftplib.FTP(self.cfg['server'])
        ftp.login(**cred)
        ftp.cwd(sim_dir)

        svar = ', '.join(str(variable_name)
                         for variable_name in variable_names)
        if not self.cfg.get('no_verbose'):
            self.log.info('{} variables for {} simulation'.format(
                svar, sim_dir))
        # Loop over all data-source variables
        for variable_name in self.variable_names:
            # Try to get variable name
            src_variable_name = self.cfg['variable_names'].get(variable_name)
            if src_variable_name is None:
                self.log.warning(
                    "{} variable not in 'variable_names' of {} configuration-"
                    "file: skipping".format(variable_name, self.name))
                continue

            # # Download all files for this variable
            # filenames = ftp.nlst(src_variable_name)
            # if src_variable_name == 'ps':
            #     # Try sea-level pressure instead of surface pressure
            #     src_variable_name = 'psl'
            #     filenames = ftp.nlst(src_variable_name)
            #     if len(filenames) == 0:
            #       if not self.cfg.get('no_verbose'):
            #           self.log.info(('\n  No file for variable {}'
            #                  'on server'.format(src_variable_name)))
            #         continue
            # else:
            #     continue

            # Loop over periods
            for start_date, end_date in zip(
                    self.start_dates, self.end_dates):
                # Get paths
                var_dir, var_file = self._get_var_dir_file(
                    variable_name=variable_name, start_date=start_date,
                    end_date=end_date)

                # Make local directories
                os.makedirs(var_dir, exist_ok=True)

                filepath = os.path.join(var_dir, var_file)
                ftp.retrbinary('RETR ' + src_variable_name + '/' + var_file,
                               open(filepath, 'wb').write)

        # Close
        ftp.close()

        return variable_names

    def load(self, transform=None, variable_names=None, **kwargs):
        """Collect all required variables from cordex simulations.

        :param transform: A function or a composee of functions
          to apply to the datasets.
          These functions should take as arguments a dataset
          and a :py:class:`.data_source.DataSourceBase` data source.
        :param variable_names: Names of variables to download.
          Default is `None`, in which case all variables in
          :py:attr:`variable_names` are downloaded.
        :type transform: :py:class:`func` or
          :py:class:`.data_source.Composer`
        :type variable_names: (collection of) :py:class:`str`

        :returns: A dataset collecting all variables and periods.
        :rtype: :py:class:`xarray.Dataset`

        .. note:: All the variables are first read, then the functions are applied
             and the different periods are finally concatenated.
        """
        # Get variable names
        variable_names = ensure_collection(
            variable_names, set) or self.variable_names

        # Collect all periods
        ds = xr.Dataset()
        read_cube, coords, src_crs, dst_crs = True, None, None, None
        for start_date, end_date in zip(
                self.start_dates, self.end_dates):
            # Collect all variables
            ds_per = xr.Dataset()
            for variable_name in variable_names:
                # Get paths
                src_variable_name = self.cfg['variable_names'].get(
                    variable_name)
                if src_variable_name is None:
                    self.log.warning(
                        "{} variable not in 'variable_names' of {} "
                        "configuration-file: skipping".format(
                            variable_name, self.name))
                    continue
                var_dir, var_file = self._get_var_dir_file(
                    variable_name=variable_name, start_date=start_date,
                    end_date=end_date)
                filepath = os.path.join(var_dir, var_file)
                if not self.cfg.get('no_verbose'):
                    self.log.info('Reading {} from {}'.format(
                        variable_name, filepath))

                # Read data
                with xr.open_dataset(filepath) as ds_var:
                    # Select variable of interest to get rid of other variables
                    da = ds_var[src_variable_name].squeeze(drop=True)

                    # Ensure that the temperature is in Kelvin
                    if 'temperature' in variable_name:
                        if da.mean() < 150.:
                            da += 273.15

                    # Ensure that the pressure is in Pa
                    if 'pressure' in variable_name:
                        da *= 10**(5 - np.round(np.log10(da.mean())))

                    # Indexes x and y may change in some data set resulting
                    # in erros when aligning. Just set with integers.
                    if 'y' in da.coords:
                        da['y'] = np.arange(da['y'].shape[0])
                    if 'x' in da.coords:
                        da['x'] = np.arange(da['x'].shape[0])
                    var = da.to_dataset(name=variable_name)

                    if self.cfg['height'] not in var[variable_name].attrs:
                        # Add height as attribute
                        if self.cfg['height'] in ds_var:
                            var[variable_name].attrs['height'] = float(
                                ds_var[self.cfg['height']])
                        elif self.cfg['height'] in ds_var.coords:
                            var[variable_name].attrs['height'] = float(
                                ds_var.coords[self.cfg['height']])
                        else:
                            # Assume 2m height
                            var[variable_name].attrs['height'] = 2.
                    else:
                        # Rename
                        var[variable_name].attrs['height'] = var[
                            variable_name].attrs[self.cfg['height']]

                    # Make sure the time indexes are exactly the same
                    # (some variables use 'days since', others
                    # 'hours from',
                    # which is not processed by python netcdf)
                    if 'time' in ds_per:
                        var['time'] = ds_per['time']

                    # Create or add variable
                    ds_per = ds_per.merge(var)

                if (read_cube and
                    (('lat' not in ds_per.coords) or
                     ('lon' not in ds_per.coords))):
                    # Get cube from first file for coordinates
                    cube = iris.load(filepath)[0]

                    # Transform to geodetic coordinates
                    coords, src_crs, dst_crs = get_geodetic_crs(cube)

                    # Prevent further cube reading
                    read_cube = False

            # Transform to geodetic coordinates
            ds_per = get_geodetic_array(ds_per, coords, src_crs, dst_crs)

            # Convert time at 12:00:00 to dates
            if self.cfg['frequency'] == 'day':
                td = pd.DatetimeIndex(ds_per.indexes['time'].date)
                ds_per.coords['time'] = td

            # Squeeze in case a one-dimensional level coordinate exists
            ds_per = ds_per.squeeze()

            # Apply functions to the dataset if given
            if transform:
                kwargs.update({'ds': ds_per, 'data_src': self})
                ds_per = transform(**kwargs)

            # Create or add period
            ds = ds.merge(ds_per)

        return ds

    def get_postfix(self, start_date=None, end_date=None, **kwargs):
        """Get standard postfix for cordex data.

        :param start_date: Simulation start-date. Default is `None`, in which
          case the first date of :py:attr::`start_dates` is used.
        :param end_date: Simulation end-date. Default is `None`, in which case
          the first date of :py:attr::`end_dates` is used.
        :type start_date: str
        :type end_date: str

        :returns: Postfix.
        :rtype: str
        """
        start_date = start_date or self.start_dates[0]
        end_date = end_date or self.end_dates[-1]

        # Get paths
        fmt = (self.cfg['domain'], self.cfg['gcm_model_name'],
               self.cfg['cmip5_experiment_name'],
               self.cfg['cmip5_ensemble_member'],
               self.cfg['rcm_model_name'], self.cfg['rcm_version_id'],
               self.cfg['frequency'], start_date, end_date)
        postfix = '_{}_{}_{}_{}_{}_{}_{}_{}-{}'.format(*fmt)

        return postfix

    def _get_sim_dir(self, **kwargs):
        """Get simulation directory of cordex data.

        returns: Source directory and filename.
        :rtype: tuple of str
        """
        sim_dir = os.path.join(
            self.cfg['domain'], self.cfg['institution'],
            self.cfg['gcm_model_name'],
            self.cfg['cmip5_experiment_name'],
            self.cfg['cmip5_ensemble_member'],
            self.cfg['rcm_model_name'],
            self.cfg['rcm_version_id'], self.cfg['frequency'])

        return sim_dir

    def get_grid(self, *args, **kwargs):
        """Get grid used to make masks.

        :returns: Dataset coordinates.
        :rtype: :py:class:`tuple` of pairs of :py:class:`str`
          and :py:class:`numpy.array`
        """
        var_dir, var_file = self._get_var_dir_file()
        filepath = os.path.join(var_dir, var_file)
        if not self.cfg.get('no_verbose'):
            self.log.info('Reading sample {} data from {}'.format(
                self.name, filepath))
        with xr.open_dataset(filepath) as ds:
            lat = ds.get('lat')
            if lat is None:
                lat = ds.get('latitude')
            lon = ds.get('lon')
            if lon is None:
                lon = ds.get('longitude')
            # Get grid coordinates
            coords = (lat.copy(deep=True).coords
                      if len(lat.shape) > 1 else
                      OrderedDict({
                          'lat': lat.copy(deep=True).values,
                          'lon': lon.copy(deep=True).values}))
        return coords

    def get_grid_postfix(self, *args, **kwargs):
        """Get grid postfix.

        :returns: Postfix.
        :rtype: str
        """
        postfix = '_{}_{}'.format(
            self.cfg['domain'], self.cfg['rcm_model_name'])

        return postfix

    def _get_var_dir_file(self, variable_name=None,
                          start_date=None, end_date=None):
        """Get variable directory and filename for cordex data.

        :param variable_name: Package name of variable for which to get
          information. Default is `None`, in which case the first
          variable in :py:attr::`variables` list is taken.
        :param start_date: Simulation start-date.
        :param end_date: Simulation end-date.
        :type variable_name: str
        :type start_date: str
        :type end_date: str


        returns: Source directory and filename.
        :rtype: tuple of str
        """
        # Take first dates if not given
        start_date = start_date or self.start_dates[0]
        end_date = end_date or self.end_dates[0]

        if variable_name is None:
            src_variable_name, it = None, 0
            while src_variable_name is None:
                variable_name = list(self.variable_names)[it]
                src_variable_name = self.cfg['variable_names'].get(
                    variable_name)
                it += 1
        else:
            src_variable_name = self.cfg['variable_names'][variable_name]

        src_dir = self.med.cfg.get_external_data_directory(self)
        sim_dir = self._get_sim_dir()
        var_dir = os.path.join(src_dir, sim_dir, src_variable_name)
        postfix = self.get_postfix(
            start_date=start_date, end_date=end_date)
        var_file = '{}{}.nc'.format(src_variable_name, postfix)

        return var_dir, var_file



"""Solar-radiation calculations."""
import numpy as np
import xarray as xr
import xarray.ufuncs as xrf
from .solar_calendar import SolarCalendarComputer, _compute_if_none
from .tracking import SolarTracker


#: Long names for horizontal irradiances.
LONG_NAMES_HORIZONTAL = {
    'global_normal_et': 'Global Normal Extraterrestrial Irradiance',
    'global_horizontal_et': 'Global Horizontal Extraterrestrial Irradiance',
    'global_horizontal_surf': 'Global Horizontal Surface Irradiance',
    'direct_normal_surf': 'Direct Normal Surface Irradiance',
    'direct_horizontal_surf': 'Direct Horizontal Surface Irradiance',
    'diffused_horizontal_surf': 'Diffuse Horizontal Surface Irradiance',
    'diffused_ratio': 'Diffuse over Global Irradiance Ratio',
    'diffused_ratio_reindl': (
        "Diffuse over Global Irradiance Ratio from Reindl's Model"),
    'clearness_index': 'Clearness Index',
}

#: Long names for tilted irradiances.
LONG_NAMES_TILTED = {
    'global_tilted_surf': 'Global Tilted Surface Irradiance',
    'direct_trans_factor': 'Direct Transposition Factor',
    'direct_tilted_surf': 'Direct Tilted Surface Irradiance',
    'diffused_trans_factor': 'Diffused Transposition Factor',
    'diffused_tilted_surf': 'Diffused Tilted Surface Irradiance',
    'view_factor': 'View Factor',
    'reflected_trans_factor': 'Reflected Transposition Factor',
    'reflected_tilted_surf': 'Reflected Tilted Surface Irradiance'
}

#: Units of irradiances.
UNITS = {
    'global_normal_et': 'W/m2',
    'global_horizontal_et': 'power',
    'clearness_index': '',
    'global_horizontal_surf': 'power',
    'global_tilted_surf': 'power',
    'direct_normal_surf': 'power',
    'direct_trans_factor': '',
    'direct_tilted_surf': 'power',
    'diffused_trans_factor': '',
    'diffused_tilted_surf': 'power',
    'view_factor': '',
    'reflected_trans_factor': '',
    'reflected_tilted_surf': 'power',
}


class HorizontalRadiationComputer():
    """Horizontal adiation computer.

    .. seealso::
       * Duffie, J.A., Beckman, W.A., 2013.
         *Solar Energy Thermal Processes*, fourth ed. Wiley, Hoboken, NJ.
       * Reindl, D.T., Beckman, W.A., Duffie, J.A., 1990.
         Diffuse Fraction Correlation. *Solar Energy* 45, 1-7.

    .. warning:: All computations are for an hourly frequency.
    """

    def __init__(
            self, time=None, lat=None, lon=None,
            global_horizontal_surf=None, global_horizontal_et=None,
            clearness_index=None, solar_constant=1367., elevation_cut=None,
            diffused_model='reindl', solar_computer=None,
            angles_in_degrees=False, **kwargs):
        """Radiation computations. All computations eventually rely on
        (i) the global horizontal surface radiation,
        :py:obj:`global_horizontal_surf`,
        and (ii) the clearness index, :py:obj:`clearness_index`.
        The clearness index may, however, be computed as the ratio of
        the global horizontal surface irradiance and the global horizontal
        extraterrestrial irradiance and the latter may be computed from
        coordinates. Thus, at least the global horizontal surface radiation
        or the clearness index should be provided.

        :param time: Time. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param lat: Latitude. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param lon: Longitude. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param global_horizontal_surf: Global horizontal surface irradiance.
          Default is `None`.
        :param global_horizontal_et: Global horizontal extraterrestrial
          irradiance. Default is `None`.
        :param clearness_index: Clearness index.
          Default is `None`.
        :param solar_constant: Solar constant (yearly averaged, in W/m2).
          Default is 1367 (W/m2).
        :param elevation_cut: Elevation threshold below which
          direct radiation is considered zero (see note).
          If `None`, use default value of 10 degrees.
        :param diffused_model: Model used to extract diffuse componant
          of global horizontal radiation. Only the.actuator by
          Reindl *et al.* (1990) is available at the moment.
          Default is `'reindl'`.
        :param solar_computer: Solar computer to initialize with, instead
          of constructing one from :py:obj:`time`, :py:obj:`lat` and
          :py:obj:`lon`, to speed-up computations.
        :param angles_in_degrees: Whether angles are given in degrees instead
          of radians and should be returned in degrees by
          :py:meth:`.solar_calendar.SolarCalendarComputer.get_angle`.
        :type time: (sequence of) datetime.
        :type lat: (sequence of) :py:class:`float`.
        :type lon: (sequence of) :py:class:`float`.
        :type global_horizontal_surf: (sequence of) :py:class:`float`.
        :type global_horizontal_et: (sequence of) :py:class:`float`.
        :type clearness_index: (sequence of) :py:class:`float`.
        :type solar_constant: float
        :type elevation_cut: float
        :type diffused_model: str
        :type solar_computer: :py:class:`.solar_calendar.SolarCalendarComputer`
        :type angles_in_degrees: bool
        """
        # Make sure that at least the global horizontal surface radiation
        # or the clearness index are provided
        if (global_horizontal_surf is None) and (clearness_index is None):
            msg = ('At least the global horizontal surface irradiance '
                   'or the clearness index should be provided.')
            raise TypeError(msg)

        # Add time, latitude and longitude (in radians) as attributes
        fact = np.deg2rad(1.) if angles_in_degrees else 1.

        # Create solar angles computer (all computations in radians)
        if solar_computer is None:
            self.time, self.lat, self.lon = time, lat * fact, lon * fact
            self.sc = SolarCalendarComputer(
                self.time, self.lat, self.lon, angles_in_degrees=False)
        else:
            self.sc = solar_computer
            self.time, self.lat, self.lon = (
                self.sc.time, self.sc.lat, self.sc.lon)

        # Initialize all attribute variables to be computed
        self.clear(LONG_NAMES_HORIZONTAL)

        # Set given radiations and and clearness index
        self._global_horizontal_surf = global_horizontal_surf
        self._global_horizontal_et = global_horizontal_et
        self._clearness_index = clearness_index

        # Get power units
        self._power_units = self._get_power_units()

        # Other parameters
        self.solar_constant = solar_constant
        self.elevation_cut = (np.deg2rad(10.) if elevation_cut is None
                              else elevation_cut * fact)
        self.diffused_model = diffused_model

    def clear(self, names):
        """Set all computed attribute variables to `None`.

        :param names: Names of variables to set.
        :type names: collection of :py:class:`str`
        """
        for variable_name in names:
            setattr(self, '_' + variable_name, None)

    def _get_power_units(self):
        """Try to get power units from
        :py:attr:`self._global_horizontal_surf` or
        :py:attr:`self._global_horizontal_et`, otherwise
        return `'power'`.

        :returns: Power units.
        :rtype: str
        """
        try:
            return self._global_horizontal_surf.attr['units']
        except (AttributeError, KeyError):
            try:
                return self._global_horizontal_et.attr['units']
            except (AttributeError, KeyError):
                return 'power'

    def _add_attributes(self, variable_name):
        """Add `long_name` and `units` attributes to data array.

        :param variable_name: Variable name.
        :type variable_name: str
        """
        # Get variable from variable name
        var = getattr(self, variable_name)

        # Try to add attributes
        try:
            variable_name = (variable_name[1:] if variable_name[0] == '_' else
                             variable_name)
            var.attrs['long_name'] = self.long_names[variable_name]
            var.attrs['units'] = (self.power_units
                                  if UNITS[variable_name] == 'power'
                                  else UNITS[variable_name])
        except AttributeError:
            pass

    @property
    @_compute_if_none
    def clearness_index(self):
        """Compute the clearness index as the ratio between the
           global (surface) and extraterrestrial horizontal radiations.

        :returns: Clearness index.
        :rtype: (sequence of) :py:class:`float`
        """
        # Compute the clearness index as the ratio between the surface and top
        # of the atmosphere downwelling shortwave radiation.
        var = xr.where(
            self.global_horizontal_et > 0.,
            self.global_horizontal_surf / self.global_horizontal_et,
            0.)

        # Ensure that the var index is between 0 and 1
        var = xr.where(var < 0., 0., var)
        var = xr.where(var > 1., 1., var)

        return var

    @property
    @_compute_if_none
    def global_normal_et(self):
        """Get extraterrestrial solar radiation for a given day of the year,
           taking into account variations or the earth-sun distance.

        :returns: Extraterrestrial solar radiation (W/m2).
        :rtype: (sequence) of float

        .. note::
          * The Fourier series.actuator of Spencer (1971) is used.
          * Hourly variations are very small and thus neglected.
        """
        doy_rad = (self.time.dayofyear - 1) * 2*np.pi / 365

        return self.solar_constant * (
            1.000110 + 0.034221 *
            xrf.cos(doy_rad) + 0.001280 * xrf.sin(doy_rad)
            + 0.000719 * xrf.cos(2*doy_rad) + 0.000077 * xrf.sin(2*doy_rad))

    @property
    @_compute_if_none
    def global_horizontal_et(self):
        """ Get the extraterrestrial horizontal radiation
            from the extraterrestrial radiation and zenith.

        :returns: Extraterrestrial horizontal radiation.
        :rtype: (sequence of) :py:class:`float`

        .. note::
          Negative values (between sunrise and sunset) are clipped to zero.
        """
        # Get the extraterrestrial horizontal radiation
        var = self.global_normal_et * xrf.cos(self.sc.zenith)

        # Keep positive values (between sunrise and sunset)
        return xr.where(var >= 0., var, 0.)

    @property
    @_compute_if_none
    def global_horizontal_surf(self):
        """ Compute the global horizontal surface radiation as the product
            of the clearness index with the global extraterrestrial
            horizontal radiation.

        :returns: Global horizontal surface radiation.
        :rtype: (sequence of) :py:class:`float`
        """
        # Get the surface horizontal radiation
        return self.clearness_index * self.global_horizontal_et

    @property
    @_compute_if_none
    def direct_normal_surf(self):
        """Get direct normal surface radiation.

        :returns: Direct normal radiation.
        :rtype: (sequence of) :py:class:`float`
        """
        return xr.where(self.direct_horizontal_surf > 1.e-10,
                        self.direct_horizontal_surf / xrf.cos(self.sc.zenith),
                        0.)

    @property
    @_compute_if_none
    def direct_horizontal_surf(self):
        """ Get the direct horizontal surface radiation.

        :returns: Direct horizontal radiation.
        :rtype: (sequence of) :py:class:`float`

        .. note:: The threshold :py:obj:`elevation_cut` should be larger than
          a few degrees in order to prevent numerical overflow in the direct
          transposition factor.

        .. seealso:: :py:meth:`TiltedRadiationComputer.direct_trans_factor`
        """
        var = self.global_horizontal_surf - self.diffused_horizontal_surf

        # Below the elevation_cut threshold, the direct radiation
        # is considered null.
        if self.elevation_cut is not None:
            var = xr.where(self.sc.elevation >= self.elevation_cut,
                           var, 0.)

        return var

    @property
    @_compute_if_none
    def diffused_horizontal_surf(self):
        """ Get diffuse component of global horizontal surface radiation.

        :returns: Diffusive part of the horizontal radiation.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.global_horizontal_surf * self.diffused_ratio

    @property
    @_compute_if_none
    def diffused_ratio(self):
        """Get ratio of diffuse over global horizontal surface radiation.

        :returns: Diffuse over global horizontal surface radiation ratio.
        :rtype: (sequence of) :py:class:`float`
        """
        if self.diffused_model == 'reindl':
            var = self.diffused_ratio_reindl

        return var

    @property
    @_compute_if_none
    def diffused_ratio_reindl(self):
        """Get ratio of diffuse over global horizontal surface radiation
        using the.actuator by Reindl * et al.* (1990).

        :returns: Diffuse over global horizontal surface radiation ratio.
        :rtype: (sequence of) :py:class:`float`
        """
        var = (1.4 - 1.749 * self.clearness_index
               + 0.177 * xrf.sin(self.sc.elevation))
        var = xr.where(
            self.clearness_index < 0.3, 1.02 - 0.254 * self.clearness_index
            + 0.0123 * xrf.sin(self.sc.elevation), var)
        var = xr.where(self.clearness_index > 0.78,
                       0.486 * self.clearness_index
                       - 0.182 * xrf.sin(self.sc.elevation), var)
        var = xr.where(var < 0.1, 0.1, var)
        var = xr.where(var > 0.97, 0.97, var)

        return var


class TiltedRadiationComputer(HorizontalRadiationComputer):
    """Tilted adiation computer.

    .. seealso::
       * Duffie, J.A., Beckman, W.A., 2013.
         *Solar Energy Thermal Processes*, fourth ed. Wiley, Hoboken, NJ.
       * Gueymard, C.A., 2009. Direct and indirect uncertainties in the
         prediction of tilted radiation for solar engineering applications.
         *Solar Energy* 83, 432-444.
       * Maleki, S.A.M, Hizam, H., Gomes, C., 2017.
         Estimation of Hourly, Daily and Monthly Global
         Solar Radiation on Inclined Surfaces: Models Re-Visited.
         *Energies* 10, 1-28.
    """

    def __init__(
            self, time=None, lat=None, lon=None, global_horizontal_surf=None,
            global_horizontal_et=None, clearness_index=None,
            solar_constant=1367., elevation_cut=0., diffused_model='reindl',
            solar_computer=None, tilt=None, azimuth=None, albedo=0.2,
            angles_in_degrees=False, **kwargs):
        """Radiation computations. All computations eventually rely on
        (i) the global horizontal surface radiation,
        :py:obj:`global_horizontal_surf`,
        and (ii) the clearness index, :py:obj:`clearness_index`.
        The clearness index may, however, be computed as the ratio of
        the global horizontal surface irradiance and the global horizontal
        extraterrestrial irradiance and the latter may be computed from
        coordinates. Thus, at least the global horizontal surface radiation
        or the clearness index should be provided.

        :param time: Time. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param lat: Latitude. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param lon: Longitude. Default is `None`, in which case
          :py:obj:`solar_computer` should be given.
        :param global_horizontal_surf: Global horizontal surface irradiance.
          Default is `None`.
        :param global_horizontal_et: Global horizontal extraterrestrial
          irradiance. Default is `None`.
        :param clearness_index: Clearness index.
          Default is `None`.
        :param solar_constant: Solar constant (yearly averaged, in W/m2).
          Default is 1367 (W/m2).
        :param elevation_cut: Elevation threshold below which
          direct radiation is considered zero (see note).
          If `None`, use default value of 10 degrees.
        :param diffused_model: Model used to extract diffuse componant
          of global horizontal radiation. Only the.actuator by
          Reindl *et al.* (1990) is available at the moment.
          Default is `'reindl'`.
        :param tilt: Surface tilt. For east-west axis, north-south axis,
          vertical axis, or two axes, respectively set to `'ew_axis'`,
          `'ns_axis'`, `'vertical'`, or `'two'`.
          For a tilt equal to the latitude, set to `'lat'`. Default is `None`,
          in which case computer may not be used to compute tilted irradiances.
        :param azimuth: Surface azimuth. Default is `None`,
          in which case computer may not be used to compute tilted irradiances.
        :param solar_constant: Solar constant (yearly averaged, in W/m2).
          Default is 1367 (W/m2).
        :param albedo: Composite ground albedo used to compute
          the reflected radiation. Default is 0.2.
        :param diffused_model: Model used to extract diffuse componant
          of global horizontal radiation. Only the.actuator by
          Reindl *et al.* (1990) is available at the moment.
          Default is `'reindl'`.
        :param solar_computer: Solar computer to initialize with, instead
          of constructing one from :py:obj:`time`, :py:obj:`lat` and
          :py:obj:`lon`, to speed-up computations.
        :param angles_in_degrees: Whether angles are given in degrees instead
          of radians.
        :type time: (sequence of) datetime.
        :type lat: (sequence of) :py:class:`float`.
        :type lon: (sequence of) :py:class:`float`.
        :type global_horizontal_surf: (sequence of) :py:class:`float`.
        :type global_horizontal_et: (sequence of) :py:class:`float`.
        :type clearness_index: (sequence of) :py:class:`float`.
        :type solar_constant: float
        :type elevation_cut: float
        :type diffused_model: str
        :type solar_computer: :py:class:`.solar_calendar.SolarCalendarComputer`
        :type tilt: (sequence of) :py:class:`float`, :py:class:`str`
        :type azimuth: (sequence of) :py:class:`float`, :py:class:`str`
        :type albedo: (sequence of) :py:class:`float`
        :type angles_in_degrees: bool
        """
        # Initialize as HorizontalRadiationComputer
        super(TiltedRadiationComputer, self).__init__(
            time=time, lat=lat, lon=lon, solar_computer=solar_computer,
            global_horizontal_surf=global_horizontal_surf,
            global_horizontal_et=global_horizontal_et,
            clearness_index=clearness_index, solar_constant=solar_constant,
            elevation_cut=elevation_cut, diffused_model=diffused_model,
            angles_in_degrees=angles_in_degrees, **kwargs)

        # Tracking computer (all computations in radians)
        self.tracker = SolarTracker(
            self.sc, tilt, azimuth,
            angles_in_degrees=angles_in_degrees, **kwargs)

        # Initialize all attribute variables to be computed
        self.clear(LONG_NAMES_TILTED)

        # Other parameters
        self.albedo = albedo

    @property
    @_compute_if_none
    def global_tilted_surf(self):
        """Compute global tilted surface radiation.

        :returns: Global tilted surface radiation.
        :rtype: (sequence of) :py:class:`float`

        .. note::
          * For a general review of the various existing.actuators to calculate
            the global tilted radiation, see Maleki *et al.* (2017)
            and Gueymard (2009).
          * For a general treatment of the available solar radiation,
            see Duffie and Beckman (2013).
        """
        # Get global radiation on tilted surface
        return (self.direct_tilted_surf + self.diffused_tilted_surf
                + self.reflected_tilted_surf)

    @property
    @_compute_if_none
    def direct_trans_factor(self):
        """Get direct transposition factor.

        :returns: Direct transposition factor.
        :rtype: (sequence of) :py:class:`float`

        .. seealso::
          :py:meth:`HorizontalRadiationComputer.direct_horizontal_surf`
        """
        # Cut at the same place as the DHI (for an elevation larger than
        # the elevation_cut, at least higher than 0)
        # and when the sun is behind is behind the surface
        test = ((self.direct_horizontal_surf > 0.)
                & (np.fabs(self.sc.azimuth - self.tracker.surface_azimuth)
                   < np.pi/2))

        return xr.where(
            test, self.tracker.cos_aoi / xrf.cos(self.sc.zenith), 0.)

    @property
    @_compute_if_none
    def direct_tilted_surf(self):
        """Get direct tilted surface radiation.

        :returns: Direct tilted surface radiation.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.direct_horizontal_surf * self.direct_trans_factor

    @property
    @_compute_if_none
    def diffused_trans_factor(self):
        """Get diffused transposition factor.

        :returns: Diffused transposition factor.
        :rtype: (sequence of) :py:class:`float`

        .. note::
          The diffuse transposition factor is calculated using the
          HDKR.actuator.
          See Reindl * et al.* (1990) and Maleki * et al.* (2017).
        """
        aniso_idx = xr.where(
            self.global_horizontal_surf > 0.,
            self.direct_horizontal_surf / self.global_horizontal_surf
            * self.clearness_index,
            0.)

        # Modulating function mod_fun
        mod_fun = xr.where(
            self.global_horizontal_surf > 0.,
            xrf.sqrt(self.direct_horizontal_surf /
                     self.global_horizontal_surf),
            0.)

        # Isotropic transposition factor
        transpo_iso = (1. + xrf.cos(self.tracker.surface_tilt)) / 2

        # Diffuse transposition factor. See note.
        return (
            aniso_idx * self.direct_trans_factor +
            (1. - aniso_idx) * transpo_iso
            * (1 + mod_fun * xrf.sin(self.tracker.surface_tilt / 2)**3))

    @property
    @_compute_if_none
    def diffused_tilted_surf(self):
        """ Get diffuse tilted surface radiation.

        :returns: Diffuse tilted surface radiation.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.diffused_horizontal_surf * self.diffused_trans_factor

    @property
    @_compute_if_none
    def view_factor(self):
        """Get view factor for reflected radiation.

        :returns: View factor.
        :rtype: (sequence of) :py:class:`float`
        """
        return (1. - xrf.cos(self.tracker.surface_tilt)) / 2

    @property
    @_compute_if_none
    def reflected_trans_factor(self):
        """Get reflected transposition factor.

        :returns: Reflected transposition factor.
        :rtype: (sequence of) :py:class:`float`
        """
        return self.albedo * self.view_factor

    @property
    @_compute_if_none
    def reflected_tilted_surf(self):
        """ Get reflected tilted surface radiation.

        :returns: Reflected tilted radiation.
        :rtype: (sequence of) :py:class:`float`

        .. note: :
           The reflected tilted radiation is calculated for a flat array
           using the isotropic sky formula reviewed in Gueymard(2009).
        """
        return self.global_horizontal_surf * self.reflected_trans_factor
